#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
import numpy as np


def plot_wigner_function(
    vals: np.ndarray,
    positions: List[List[float]],
    momentums: List[List[float]],
    levels: int = 50,
) -> None:
    """
    Plots the Wigner function using matplotlib.

    Args:
        vals: Values to plot.
        positions: List of lists of positions.
        momentums: List of lists of momentums.
        levels: Number of contour levels.
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("The visualization feature requires matplotlib.")

    plt.contourf(positions, momentums, vals, levels=levels, cmap="RdBu")
    plt.colorbar()
    plt.xlabel("Position")
    plt.ylabel("Momentum")
    plt.title("Wigner Function")
    plt.show()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq


def test_FockState_reduced():
    with pq.Program() as program:
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1)) / 4

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2)) / 4
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(2, 0)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(2, 0)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(0, 2)) * np.sqrt(1 / 8)

    simulator = pq.FockSimulator(d=2)
    state = simulator.execute(program).state

    with pq.Program() as reduced_program:
        pq.Q() | pq.DensityMatrix(ket=(1,), bra=(1,)) / 4

        pq.Q() | pq.DensityMatrix(ket=(2,), bra=(2,)) / 4
        pq.Q() | pq.DensityMatrix(ket=(0,), bra=(0,)) / 2

    reduced_program_simulator = pq.FockSimulator(d=1)
    reduced_program_state = reduced_program_simulator.execute(reduced_program).state

    expected_reduced_state = reduced_program_state

    reduced_state = state.reduced(modes=(1,))

    assert expected_reduced_state == reduced_state


def test_FockState_fock_amplitudes_map():

    theta = 0.3
    with pq.Program() as program:
        pq.Q() | pq.DensityMatrix(ket=(1, 0), bra=(0, 1)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2)) / 4
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(2, 0)) / 2

        pq.Q(0) | pq.Phaseshifter(theta)

    simulator = pq.FockSimulator(d=2)
    state = simulator.execute(program).state

    expected_fock_amplitudes = {
        ((1, 0), (0, 1)): 1 / 2 * np.exp(1j * theta),
        ((2, 0), (2, 0)): 0.5 + 0.0j,
        ((0, 2), (0, 2)): 1 / 4,
    }

    actual_fock_amplitudes = state.fock_amplitudes_map

    assert len(actual_fock_amplitudes.items()) == len(expected_fock_amplitudes.items())

    for occupation_number, expected_ampl in expected_fock_amplitudes.items():
        assert np.isclose(
            actual_fock_amplitudes[occupation_number],
            expected_ampl,
        )

    # Try some Fock states that are expected to have zero amplitudes
    assert np.isclose(actual_fock_amplitudes[(0, 0)], 0.0)
    assert np.isclose(actual_fock_amplitudes[(1, 0)], 0.0)
    assert np.isclose(actual_fock_amplitudes[(2, 1)], 0.0)
    assert np.isclose(actual_fock_amplitudes[(122, 35)], 0.0)


def test_FockState_fock_probabilities_map():
    with pq.Program() as program:
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1)) / 4

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2)) / 4
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(2, 0)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(2, 0)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(0, 2)) * np.sqrt(1 / 8)

    simulator = pq.FockSimulator(d=2)
    state = simulator.execute(program).state

    expected_fock_probabilities = {
        (0, 0): 0.0,
        (0, 1): 0.25,
        (1, 0): 0.0,
        (0, 2): 0.25,
        (1, 1): 0.0,
        (2, 0): 0.5,
        (0, 3): 0.0,
        (1, 2): 0.0,
        (2, 1): 0.0,
        (3, 0): 0.0,
    }

    actual_fock_probabilities = state.fock_probabilities_map

    for occupation_number, expected_probability in expected_fock_probabilities.items():
        assert np.isclose(
            actual_fock_probabilities[occupation_number],
            expected_probability,
        )


def test_FockState_get_marginal_fock_probabilities():
    with pq.Program() as program:
        pq.Q() | pq.DensityMatrix(ket=(0, 1), bra=(0, 1)) / 4

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(0, 2)) / 4
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(2, 0)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 2), bra=(2, 0)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(2, 0), bra=(0, 2)) * np.sqrt(1 / 8)

    simulator = pq.FockSimulator(d=2)
    state = simulator.execute(program).state

    modes = (1,)
    expected_probabilities = np.array([0.5, 0.25, 0.25, 0.0])
    actual_probabilities = state.get_marginal_fock_probabilities(modes)

    assert np.allclose(actual_probabilities, expected_probabilities)


def test_FockState_quadratures_mean_variance():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=0.2, phi=0)
        pq.Q(1) | pq.Squeezing(r=0.1, phi=1.0)
        pq.Q(2) | pq.Displacement(r=0.2, phi=np.pi / 2)

    simulator = pq.FockSimulator(d=3, config=pq.Config(cutoff=8, hbar=2))
    result = simulator.execute(program)

    mean_on_0th, variance_on_0th = result.state.quadratures_mean_variance(modes=(0,))
    mean_on_1st, variance_on_1st = result.state.quadratures_mean_variance(modes=(1,))
    mean_on_2nd, variance_on_2nd = result.state.quadratures_mean_variance(
        modes=(0,), phi=np.pi / 2
    )
    mean_on_3rd, variance_on_3rd = result.state.quadratures_mean_variance(
        modes=(2,), phi=np.pi / 2
    )

    assert np.isclose(mean_on_0th, 0.4, rtol=1e-5)
    assert mean_on_1st == 0.0
    assert np.isclose(mean_on_2nd, 0.0)
    assert np.isclose(mean_on_3rd, 0.4, rtol=1e-5)
    assert np.isclose(variance_on_0th, 1.0, rtol=1e-5)
    assert np.isclose(variance_on_1st, 0.9112844, rtol=1e-5)
    assert np.isclose(variance_on_2nd, 1, rtol=1e-5)
    assert np.isclose(variance_on_3rd, 1, rtol=1e-5)


def test_FockState_non_selfadjoint_density_matrix_raises_InvalidState():

    state = pq.FockState(d=1, connector=pq.NumpyConnector())

    non_selfadjoint_matrix = np.array([[1, 2], [3, 4]])

    state._density_matrix = non_selfadjoint_matrix

    with pytest.raises(pq.api.exceptions.InvalidState) as error:
        state.validate()

    assert "The density matrix is not self-adjoint" in error.value.args[0]


def test_FockState_non_selfadjoint_density_matrix_no_error_if_validate_False():

    state = pq.FockState(
        d=1, connector=pq.NumpyConnector(), config=pq.Config(validate=False)
    )

    non_selfadjoint_matrix = np.array([[1, 2], [3, 4]])

    state._density_matrix = non_selfadjoint_matrix

    state.validate()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import piquasso as pq
import numpy as np


def test_PostSelectPhotons_gradient(tf):
    def _calculate_loss(weights, connector, state_vector):
        np = connector.np

        with pq.Program() as preparation:
            pq.Q(all) | pq.StateVector([0, 1, 0]) * state_vector[0]
            pq.Q(all) | pq.StateVector([1, 1, 0]) * state_vector[1]
            pq.Q(all) | pq.StateVector([2, 1, 0]) * state_vector[2]

        phase_shifter_phis = weights[:3]
        thetas = weights[3:6]
        phis = weights[6:]
        with pq.Program() as interferometer:
            for i in range(3):
                pq.Q(i) | pq.Phaseshifter(phase_shifter_phis[i])

            pq.Q(1, 2) | pq.Beamsplitter(theta=thetas[0], phi=phis[0])
            pq.Q(0, 1) | pq.Beamsplitter(theta=thetas[1], phi=phis[1])
            pq.Q(1, 2) | pq.Beamsplitter(theta=thetas[2], phi=phis[2])

        with pq.Program() as program:
            pq.Q(all) | preparation

            pq.Q(all) | interferometer

            pq.Q(all) | pq.PostSelectPhotons(
                postselect_modes=(1, 2), photon_counts=(1, 0)
            )

        simulator = pq.PureFockSimulator(
            d=3, config=pq.Config(cutoff=4), connector=connector
        )

        state = simulator.execute(program).state

        state.normalize()

        density_matrix = state.density_matrix[:3, :3]

        expected_state = np.copy(state_vector)
        expected_state = connector.assign(expected_state, 2, -expected_state[2])

        loss = 1 - np.sqrt(
            np.real(np.conj(expected_state) @ density_matrix @ expected_state)
        )

        return loss

    connector = pq.TensorflowConnector()

    weights = tf.Variable(
        [np.pi, 0.0, 0.0, np.pi / 8, 65.5302 * 2 * np.pi / 360, -np.pi / 8, 0, 0, 0]
    )

    with tf.GradientTape() as tape:
        loss = _calculate_loss(
            weights=weights,
            connector=connector,
            state_vector=np.sqrt([0.2, 0.3, 0.5]),
        )

    grad = tape.gradient(loss, weights)

    assert np.isclose(loss, 0.0)

    assert np.allclose(grad, 0.0, atol=1e-7)


def test_ImperfectPostSelectPhotons_gradient(tf):
    def _calculate_loss(weights, connector, state_vector):
        np = connector.np

        with pq.Program() as preparation:
            pq.Q(all) | pq.StateVector([0, 1, 0]) * state_vector[0]
            pq.Q(all) | pq.StateVector([1, 1, 0]) * state_vector[1]
            pq.Q(all) | pq.StateVector([2, 1, 0]) * state_vector[2]

        phase_shifter_phis = weights[:3]
        thetas = weights[3:6]
        phis = weights[6:]
        with pq.Program() as interferometer:
            for i in range(3):
                pq.Q(i) | pq.Phaseshifter(phase_shifter_phis[i])

            pq.Q(1, 2) | pq.Beamsplitter(theta=thetas[0], phi=phis[0])
            pq.Q(0, 1) | pq.Beamsplitter(theta=thetas[1], phi=phis[1])
            pq.Q(1, 2) | pq.Beamsplitter(theta=thetas[2], phi=phis[2])

        with pq.Program() as program:
            pq.Q(all) | preparation

            pq.Q(all) | interferometer

            pq.Q(all) | pq.ImperfectPostSelectPhotons(
                postselect_modes=(1, 2),
                photon_counts=(1, 0),
                detector_efficiency_matrix=np.array(
                    [
                        [1.0, 0.2, 0.1],
                        [0.0, 0.8, 0.2],
                        [0.0, 0.0, 0.7],
                    ]
                ),
            )

        simulator = pq.PureFockSimulator(
            d=3, config=pq.Config(cutoff=4), connector=connector
        )

        state = simulator.execute(program).state

        state.normalize()

        density_matrix = state.density_matrix[:3, :3]

        expected_state = np.copy(state_vector)
        expected_state = connector.assign(expected_state, 2, -expected_state[2])

        loss = 1 - np.sqrt(
            np.real(np.conj(expected_state) @ density_matrix @ expected_state)
        )

        return loss

    connector = pq.TensorflowConnector()

    weights = tf.Variable(
        [np.pi, 0.0, 0.0, np.pi / 8, 65.5302 * 2 * np.pi / 360, -np.pi / 8, 0, 0, 0]
    )

    with tf.GradientTape() as tape:
        loss = _calculate_loss(
            weights=weights,
            connector=connector,
            state_vector=np.sqrt([0.2, 0.3, 0.5]),
        )

    grad = tape.gradient(loss, weights)

    assert np.isclose(loss, 0.0485518312033707)

    assert np.allclose(
        grad,
        [
            0.0,
            0.0,
            0.0,
            0.08767547,
            0.01909362,
            -0.01343052,
            0.0,
            0.0,
            0.0,
        ],
        atol=1e-7,
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from piquasso.api.result import Result

from piquasso.api.exceptions import InvalidParameter

from piquasso._math.validations import all_zero_or_one, are_modes_consecutive

from ._utils import (
    calculate_indices_for_controlled_phase,
    calculate_indices_for_ising_XX,
)

from .._utils import (
    get_fock_space_index,
    get_cutoff_fock_space_dimension,
    next_second_quantized,
)
from .state import PureFockState

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from piquasso.instructions.preparations import StateVector
    from piquasso.instructions.gates import _PassiveLinearGate, Squeezing2
    from piquasso.fermionic.instructions import ControlledPhase, IsingXX


def state_vector(
    state: PureFockState, instruction: "StateVector", shots: int
) -> Result:
    connector = state._connector

    fallback_np = connector.fallback_np

    coefficient = instruction._all_params["coefficient"]

    if "occupation_numbers" in instruction._all_params:
        occupation_numbers = fallback_np.array(
            instruction._all_params["occupation_numbers"]
        )

        if state._config.validate and not all_zero_or_one(occupation_numbers):
            raise InvalidParameter(
                f"Invalid initial state specified: instruction={instruction}"
            )

        index = get_fock_space_index(occupation_numbers)

        state._state_vector = connector.assign(state._state_vector, index, coefficient)

    elif "fock_amplitude_map" in instruction._all_params:
        for occupation, amplitude in instruction._all_params[
            "fock_amplitude_map"
        ].items():
            occ_numbers = fallback_np.array(occupation)

            if state._config.validate and (
                len(occ_numbers) != state._d or not all_zero_or_one(occ_numbers)
            ):
                raise InvalidParameter(
                    f"Invalid initial state specified: instruction={instruction}"
                )

            index = get_fock_space_index(occ_numbers)

            state._state_vector = connector.assign(
                state._state_vector,
                index,
                coefficient * amplitude,
            )

    return Result(state=state)


def passive_linear(
    state: PureFockState, instruction: "_PassiveLinearGate", shots: int
) -> Result:
    connector = state._connector
    config = state._config

    modes = instruction.modes

    if config.validate and not are_modes_consecutive(modes):
        raise InvalidParameter(f"Specified modes must be consecutive: modes={modes}")

    unitary = instruction._get_passive_block(connector, config)

    representations = connector.calculate_interferometer_on_fermionic_fock_space(
        unitary, config.cutoff
    )

    state._state_vector = connector.apply_fermionic_passive_linear_to_state_vector(
        representations,
        state._state_vector,
        modes,
        state._d,
        config.cutoff,
    )

    return Result(state=state)


def squeezing2(state: PureFockState, instruction: "Squeezing2", shots: int) -> Result:
    connector = state._connector
    config = state._config

    modes = instruction.modes

    d = state._d
    cutoff = state._config.cutoff

    np = connector.np
    fallback_np = connector.fallback_np

    if config.validate and not are_modes_consecutive(modes):
        raise InvalidParameter(f"Specified modes must be consecutive: modes={modes}")

    r = instruction.params["r"]
    phi = instruction.params["phi"]

    size = get_cutoff_fock_space_dimension(d, cutoff)

    index = fallback_np.zeros(d, dtype=np.int64)

    U = np.array(
        [
            [np.cos(r / 2), np.sin(r / 2) * np.exp(-1j * phi)],
            [-np.sin(r / 2) * np.exp(1j * phi), np.cos(r / 2)],
        ]
    )

    for i in range(size):
        if index[modes[0]] == 0 and index[modes[1]] == 0:
            index[modes[0]] = 1
            index[modes[1]] = 1
            j = get_fock_space_index(index)
            index[modes[0]] = 0
            index[modes[1]] = 0

            if j < size:
                state._state_vector = connector.assign(
                    state._state_vector, ((i, j),), U @ state._state_vector[(i, j),]
                )
            else:
                state._state_vector = connector.assign(
                    state._state_vector, i, U[0, 0] * state._state_vector[i]
                )

        index = next_second_quantized(index)

    return Result(state=state)


def controlled_phase(
    state: PureFockState, instruction: "ControlledPhase", shots: int
) -> Result:
    connector = state._connector

    modes = instruction.modes

    d = state._d
    cutoff = state._config.cutoff

    np = connector.np

    phi = instruction.params["phi"]

    rotation = np.exp(1j * phi)

    indices = calculate_indices_for_controlled_phase(d, cutoff, modes)

    state._state_vector = connector.assign(
        state._state_vector, indices, rotation * state._state_vector[indices]
    )

    return Result(state=state)


def ising_XX(state: PureFockState, instruction: "IsingXX", shots: int) -> Result:
    connector = state._connector
    np = connector.np

    phi = instruction.params["phi"]
    modes = instruction.modes

    d = state._d
    cutoff = state._config.cutoff

    cos_phi = np.cos(phi)
    i_sin_phi = 1j * np.sin(phi)

    indices = calculate_indices_for_ising_XX(d, cutoff, modes)

    for index in indices:
        initial = state._state_vector[index]
        final = cos_phi * initial + i_sin_phi * np.flip(initial)
        state._state_vector = connector.assign(state._state_vector, index, final)

    return Result(state=state)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from piquasso.api.result import Result
from piquasso.api.exceptions import InvalidParameter
from piquasso.api.instruction import Instruction

from piquasso.instructions.gates import _PassiveLinearGate, Squeezing2
from piquasso.fermionic.instructions import IsingXX

from piquasso._math.validations import all_zero_or_one
from piquasso._math.transformations import (
    xxpp_to_xpxp_indices,
)

from piquasso._math.indices import double_modes

from .._utils import validate_fermionic_gaussian_hamiltonian
from .state import GaussianState

import typing

if typing.TYPE_CHECKING:
    import numpy as np
    from typing import Tuple


def vacuum(state: GaussianState, instruction: Instruction, shots: int) -> Result:
    """Sets the vacuum state."""

    np = state._connector.np

    state._set_occupation_numbers(np.zeros(state.d, dtype=state._config.dtype))

    return Result(state=state)


def state_vector(state: GaussianState, instruction: Instruction, shots: int) -> Result:
    """Sets an arbitrary occupation number state."""

    fallback_np = state._connector.fallback_np

    coefficient = instruction._all_params["coefficient"]
    if state._config.validate and not fallback_np.isclose(coefficient, 1.0):
        raise InvalidParameter(
            "Only 1.0 is permitted as coefficient for the state vector."
        )

    occupation_numbers = fallback_np.array(
        instruction._all_params["occupation_numbers"]
    )

    if state._config.validate and not all_zero_or_one(occupation_numbers):
        raise InvalidParameter(
            f"Invalid initial state specified: instruction={instruction}"
        )

    state._set_occupation_numbers(occupation_numbers)

    return Result(state=state)


def parent_hamiltonian(
    state: GaussianState, instruction: Instruction, shots: int
) -> Result:
    state._set_parent_hamiltonian(instruction.params["hamiltonian"])

    return Result(state=state)


def gaussian_hamiltonian(
    state: GaussianState, instruction: Instruction, shots: int
) -> Result:
    hamiltonian = instruction.params["hamiltonian"]
    modes = instruction.modes

    validate_fermionic_gaussian_hamiltonian(hamiltonian)
    h = _transform_to_majorana_basis(hamiltonian, state._connector)

    evolved_state = _do_apply_gaussian_hamiltonian(state, h, modes)

    return Result(state=evolved_state)


def _transform_to_majorana_basis(H, connector):
    r"""Transform quadratic gate Hamiltonian to Majorana basis.

    Consider a gate unitary of the form

    .. math::
        U = e^{i \hat{H}},

    where

    .. math::
        \hat{H} &= \mathbf{f} H \mathbf{f}^\dagger, \\\\
        H &= \begin{bmatrix}
            A & -\overline{B} \\
            B & -\overline{A}
        \end{bmatrix}.

    Changing to Majorana basis, one can rewrite this as

    .. math::
        \hat{H} &= i \sum_{j,k=1}^{2d} h_{jk} m_j m_k,

    where :math:`m_j` denotes the Majorana operators

    .. math::
        m_{2k} &:= f_k + f_k^\dagger, \\\\
        p_{2k+1} &:= -i (f_k - f_k^\dagger).

    Returns:
        np.ndarray: The matrix `h`.
    """

    np = connector.np
    fallback_np = connector.fallback_np

    small_d = len(H) // 2

    A = H[small_d:, small_d:]
    B = H[:small_d, small_d:]

    A_plus_B = A + B
    A_minus_B = A - B
    indices = xxpp_to_xpxp_indices(small_d)

    return (
        np.block(
            [
                [A_plus_B.imag, A_plus_B.real],
                [-A_minus_B.real, A_minus_B.imag],
            ]
        )[fallback_np.ix_(indices, indices)]
        / 2
    )


def _do_apply_gaussian_hamiltonian(
    state: GaussianState, h: "np.ndarray", modes: "Tuple[int, ...]"
) -> GaussianState:
    connector = state._connector
    np = connector.np
    fallback_np = connector.fallback_np

    d = state._d

    doubled_modes = double_modes(fallback_np.array(modes))

    SO = connector.embed_in_identity(
        connector.expm(-4 * h), np.ix_(doubled_modes, doubled_modes), 2 * d
    )

    state.covariance_matrix = SO @ state.covariance_matrix @ SO.T

    return state


def passive_linear_gate(
    state: GaussianState, instruction: _PassiveLinearGate, shots: int
) -> Result:
    r"""Applies a passive linear gate to a fermionic Gaussian state.

    The transformation is equivalent to

    .. math::
        \Gamma^{a^\dagger a} &\mapsto \overline{U} \Gamma^{a^\dagger a} U^T, \\\\
        \Gamma^{a^\dagger a^\dagger} &\mapsto \overline{U} \Gamma^{a^\dagger a^\dagger} U^\dagger,

    where :math:`U` is the unitary corresponding to the passive linear gate in the Dirac
    basis.
    """  # noqa: E501

    unitary = instruction._get_passive_block(state._connector, state._config)
    modes = instruction.modes

    d = state._d

    connector = state._connector

    fallback_np = connector.fallback_np

    all_modes = fallback_np.arange(d)

    select_columns = fallback_np.ix_(all_modes, modes)
    select_rows = fallback_np.ix_(modes, all_modes)

    state._D = connector.assign(
        state._D, select_columns, state._D[select_columns] @ unitary.T
    )
    state._D = connector.assign(
        state._D, select_rows, unitary.conj() @ state._D[modes, :]
    )

    state._E = connector.assign(
        state._E, select_columns, state._E[select_columns] @ unitary.T.conj()
    )
    state._E = connector.assign(
        state._E, select_rows, unitary.conj() @ state._E[select_rows]
    )

    return Result(state=state)


def squeezing2(state: GaussianState, instruction: Squeezing2, shots: int) -> Result:
    connector = state._connector
    np = connector.np
    complex_dtype = state._D.dtype

    modes = instruction.modes
    r = instruction.params["r"]
    phi = instruction.params["phi"]

    r_cos_phi = r * np.cos(phi)
    r_sin_phi = r * np.sin(phi)

    h_active = (
        np.array(
            [
                [r_cos_phi, r_sin_phi],
                [r_sin_phi, -r_cos_phi],
            ],
            dtype=complex_dtype,
        )
        / 8
    )

    zeros = np.zeros_like(h_active)

    h = np.block([[zeros, h_active], [-h_active, zeros]])

    evolved_state = _do_apply_gaussian_hamiltonian(state, h, modes)

    return Result(state=evolved_state)


def ising_XX(state: GaussianState, instruction: IsingXX, shots: int) -> Result:
    phi = instruction.params["phi"]

    connector = state._connector

    np = connector.np

    modes = instruction.modes

    h = np.zeros((4, 4), dtype=state._config.dtype)

    h = connector.assign(h, (1, 2), -phi / 2)
    h = connector.assign(h, (2, 1), phi / 2)

    evolved_state = _do_apply_gaussian_hamiltonian(state, h, modes)

    return Result(state=evolved_state)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from functools import partial

from ..connector import BuiltinConnector


class TensorflowConnector(BuiltinConnector):
    """Connector enabling calculating the gradients of certain instructions.

    This connector is similar to
    :class:`~piquasso._simulators.connector.NumpyConnector`, but it enables the
    simulator to use Tensorflow to be able to compute gradients.

    Example usage::

        import piquasso as pq
        import tensorflow as tf

        r = tf.Variable(0.43)

        tensorflow_connector = pq.TensorflowConnector()

        simulator = pq.PureFockSimulator(d=1, connector=tensorflow_connector)

        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Displacement(r=r)

        with tf.GradientTape() as tape:
            state = simulator.execute(program).state

            mean = state.mean_photon_number()

        gradient = tape.gradient(mean, [r])

    Note:
        Non-deterministic operations like
        :class:`~piquasso.instructions.measurements.ParticleNumberMeasurement` are
        non-differentiable, please use a deterministic attribute of the resulting state
        instead.

    """

    def __init__(self, decorate_with=None):
        """
        Args:
            decorate_with (function, optional): A function to decorate calculations
                with. Currently, only `tf.function` is supported. Specifying this may
                reduce runtime after the tracing step. See
                `Better performance with tf.function <https://www.tensorflow.org/guide/function>`_.

        Raises:
            ImportError: When TensorFlow is not available.
        """  # noqa: E501
        try:
            import tensorflow as tf
        except ImportError:
            raise ImportError(
                "You have invoked a feature which requires 'tensorflow'.\n"
                "You can install tensorflow via:\n"
                "\n"
                "pip install piquasso[tensorflow]"
            )

        import tensorflow.experimental.numpy as tnp
        from tensorflow.python.ops.numpy_ops import np_config

        np_config.enable_numpy_behavior()

        self._tf = tf
        self.np = tnp
        self.fallback_np = np

        self._decorate_with = decorate_with

        # NOTE: `_decorated_functions` are not copied in `__deepcopy__`, but it is not
        # necessarily a problem, worst case scenario is a retracing. However, I think
        # there would be no problem copying `_decorated_functions` as long as the
        # decorated functions do not depend on local variables outside of the function.
        # Not doing this, retracing may occur when providing an initial state in
        # `Program.execute_instructions`. Solving this might be better to avoid using
        # `deepcopy` and implement `copy` functions for each `State` subclass.
        self._decorated_functions = dict()

        self._decorator_provided = self._decorate_with is not None

        if self._decorator_provided:
            self.decorator = self._decorator

        self.sqrtm = tf.linalg.sqrtm

        self.range = tf.range

    @property
    def _custom_gradient_enabled(self):
        return not self._decorator_provided and self._tf.executing_eagerly()

    @property
    def forward_pass_np(self):
        return self.fallback_np if self._custom_gradient_enabled else self.np

    @property
    def custom_gradient(self):
        return (
            self._tf.custom_gradient
            if self._custom_gradient_enabled
            else super().custom_gradient
        )

    def _decorator(self, func):
        if func.__name__ in self._decorated_functions:
            return self._decorated_functions[func.__name__]

        decorated = self._decorate_with(func)

        self._decorated_functions[func.__name__] = decorated

        return decorated

    def preprocess_input_for_custom_gradient(self, value):
        if not self._custom_gradient_enabled:
            return value

        return value.numpy() if self._tf.is_tensor(value) else value

    def block_diag(self, *arrs):
        block_diagonalized = self._tf.linalg.LinearOperatorBlockDiag(
            [self._tf.linalg.LinearOperatorFullMatrix(arr) for arr in arrs]
        )

        return block_diagonalized.to_dense()

    def assign(self, array, index, value):
        # NOTE: This method is very limited, and is a bit hacky, since TF does not
        # support item assignment through its NumPy API.

        if isinstance(array, self.fallback_np.ndarray):
            array[index] = value

            return array

        if isinstance(index, int):
            return self._tf.tensor_scatter_nd_update(array, [[index]], [value])

        # NOTE: When using `tf.function`, TensorFlow threw the following error:
        #
        # TypeError: Tensors in list passed to 'values' of 'ConcatV2' Op have types [int32, int64] that don't all match.  # noqa: E501
        #
        # To make it disappear, I had to convert all the indices to `int32`.
        index = index.astype(self.fallback_np.int32)

        if len(array.shape) == 1:
            return self._tf.tensor_scatter_nd_update(
                array, index.reshape(-1, 1), value.reshape(-1)
            )

        number_of_batches = array.shape[1]
        int_dtype = index.dtype

        flattened_index = index.reshape(-1)

        indices = self.fallback_np.column_stack(
            [
                self.fallback_np.tile(flattened_index, number_of_batches),
                self.fallback_np.concatenate(
                    [
                        self.fallback_np.full(len(flattened_index), i, dtype=int_dtype)
                        for i in range(number_of_batches)
                    ]
                ),
            ]
        )

        values = self.np.concatenate(
            [value[:, :, i].reshape(-1) for i in range(number_of_batches)]
        )

        return self._tf.tensor_scatter_nd_update(
            array,
            indices,
            values,
        )

    def block(self, arrays):
        # NOTE: This is not as advanced as `numpy.block`, this function only supports
        # 4 same-length blocks.

        return self._tf.concat(
            [self._tf.concat(arrays[0], 1), self._tf.concat(arrays[1], 1)], 0
        )

    def scatter(self, indices, updates, shape):
        return self._tf.scatter_nd(indices, updates, shape)

    def embed_in_identity(self, matrix, indices, dim):
        tf_indices = []
        updates = []

        small_dim = len(indices[0])
        for row in range(small_dim):
            for col in range(small_dim):
                index = [indices[0][row][col], indices[1][row][col]]
                update = matrix[row, col]

                tf_indices.append(index)
                updates.append(update)

        for index in range(dim):
            diagonal_index = [index, index]
            if diagonal_index not in tf_indices:
                tf_indices.append(diagonal_index)
                updates.append(1.0)

        return self.scatter(tf_indices, updates, (dim, dim))

    def _funm(self, matrix, func):
        eigenvalues, U = self._tf.linalg.eig(matrix)

        log_eigenvalues = func(eigenvalues)

        return U @ self.np.diag(log_eigenvalues) @ self._tf.linalg.inv(U)

    def logm(self, matrix):
        # NOTE: Tensorflow 2.0 has matrix logarithm, but it doesn't support gradient.
        # Therefore we had to implement our own.
        return self._funm(matrix, self.np.log)

    def expm(self, matrix):
        # NOTE: Tensorflow 2.0 has matrix exponential, but it doesn't support gradient.
        # Therefore we had to implement our own.
        return self._funm(matrix, self.np.exp)

    def powm(self, matrix, power):
        return self._funm(matrix, partial(self.np.power, x2=power))

    def polar(self, matrix, side="right"):
        P = self._tf.linalg.sqrtm(self.np.conj(matrix) @ matrix.T)
        Pinv = self._tf.linalg.inv(P)

        if side == "right":
            U = matrix @ Pinv
        elif side == "left":
            U = Pinv @ matrix

        return U, P

    def svd(self, matrix):
        # NOTE: Tensorflow 2.0 SVD has different return tuple.

        S, V, W = self._tf.linalg.svd(matrix)

        return V, S, self.np.conj(W).T

    def gather_along_axis_1(self, array, indices):
        # NOTE: Gather along axis 1 was terribly slow in Tensorflow, see
        # https://github.com/tensorflow/ranking/issues/160.

        np = self.fallback_np

        size = array.shape[0]

        size_range = np.arange(size)

        reshaped_indices = []

        for row in size_range:
            reshaped_indices.append(
                np.stack([np.full(indices.shape, row), indices], axis=2)
            )

        return self._tf.gather_nd(array, np.array(reshaped_indices))

    def transpose(self, matrix):
        # NOTE: Similarly to `tf.gather(..., axis=1)`, `tf.transpose` is also pretty
        # slow when JIT compiled, and its einsum implementation is somehow faster.

        return self.np.einsum("ij->ji", matrix)

    def accumulator(self, dtype, size, **kwargs):
        if self._custom_gradient_enabled:
            return super().accumulator(dtype, size, **kwargs)

        return self._tf.TensorArray(dtype=dtype, size=size, **kwargs)

    def write_to_accumulator(self, accumulator, index, value):
        if self._custom_gradient_enabled:
            return super().write_to_accumulator(accumulator, index, value)

        return accumulator.write(index, value)

    def stack_accumulator(self, accumulator):
        if self._custom_gradient_enabled:
            return super().stack_accumulator(accumulator)

        return accumulator.stack()

    def __tf_tracing_type__(self, context):
        # NOTE: We need to create a `TraceType` for `TensorflowConnector` to avoid
        # retracing, but it cannot be defined on module level due to the dependence
        # on `tensorflow`.
        # See `https://www.tensorflow.org/guide/function#use_the_tracing_protocol`_.

        class _TrivialTraceType(self._tf.types.experimental.TraceType):
            def __init__(self, connector):
                self.connector = connector

            def is_subtype_of(self, other):
                return True

            def most_specific_common_supertype(self, others):
                return self

            def placeholder_value(self, placeholder_context=None):
                return self.connector

            def __eq__(self, other):
                return True

            def __hash__(self):
                return 1

        return _TrivialTraceType(self)

    def permanent(self, matrix, rows, columns):
        raise NotImplementedError()

    def hafnian(self, matrix, reduce_on):
        raise NotImplementedError()

    def loop_hafnian(self, matrix, diagonal, reduce_on):
        raise NotImplementedError()

    def loop_hafnian_batch(self, matrix, diagonal, reduce_on, cutoff):
        raise NotImplementedError()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import piquasso as pq
from piquasso.decompositions.clements import clements

from scipy.stats import unitary_group

pytestmark = pytest.mark.benchmark(
    group="pure-fock-clements-comparison",
)


@pytest.fixture
def cutoff():
    return 5


d_tuple = (2, 3, 4, 5)
U_tuple = tuple([unitary_group.rvs(d) for d in d_tuple])


@pytest.mark.parametrize("d, U", zip(d_tuple, U_tuple))
def piquasso_interferometer_benchmark(benchmark, d, cutoff, U):
    @benchmark
    def func():
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            for i in range(d):
                pq.Q(i) | pq.Squeezing(r=0.1)

            pq.Q() | pq.Interferometer(U)

        simulator_fock = pq.PureFockSimulator(d=d, config=pq.Config(cutoff=cutoff))

        simulator_fock.execute(program)


@pytest.mark.parametrize("d, U", zip(d_tuple, U_tuple))
def piquasso_clements_benchmark(benchmark, d, cutoff, U):
    decomposition = clements(U, connector=pq.NumpyConnector())

    @benchmark
    def func():
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            for i in range(d):
                pq.Q(i) | pq.Squeezing(r=0.1)

            for operation in decomposition.beamsplitters:
                pq.Q(operation.modes[0]) | pq.Phaseshifter(phi=operation.params[1])
                pq.Q(*operation.modes) | pq.Beamsplitter(operation.params[0], 0.0)

            for operation in decomposition.phaseshifters:
                pq.Q(operation.mode) | pq.Phaseshifter(operation.phi)

        simulator_fock = pq.PureFockSimulator(d=d, config=pq.Config(cutoff=cutoff))

        simulator_fock.execute(program)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import piquasso as pq
import tensorflow as tf

import time


tf.get_logger().setLevel("ERROR")


def measure_graph_size(f, *args):
    if not hasattr(f, "get_concrete_function"):
        return 0

    g = f.get_concrete_function(*args).graph

    return len(g.as_graph_def().node)


def calculate_mean_position(weights, cutoff, d, connector):
    simulator = pq.PureFockSimulator(
        d,
        pq.Config(cutoff=cutoff),
        connector=connector,
    )

    with tf.GradientTape() as tape:
        cvqnn_layers = pq.cvqnn.create_layers(weights)

        preparation = [pq.Vacuum()]

        program = pq.Program(instructions=preparation + cvqnn_layers.instructions)

        final_state = simulator.execute(program).state

        mean_position = final_state.mean_position(0)

    print("_FORWARD FINISH:", time.time() - start_time)

    mean_position_grad = tape.gradient(mean_position, weights)

    print("_BACK FINISH:", time.time() - start_time)

    return mean_position, mean_position_grad


if __name__ == "__main__":
    d = 2
    layer_count = 5
    cutoff = 10

    NUMBER_OF_ITERATIONS = 5

    weights = tf.Variable(
        pq.cvqnn.generate_random_cvqnn_weights(layer_count=layer_count, d=d)
    )

    decorator = tf.function(jit_compile=True)

    connector = pq.TensorflowConnector(decorate_with=decorator)

    enhanced_calculate_mean_position = decorator(calculate_mean_position)

    print("START")
    start_time = time.time()

    enhanced_calculate_mean_position(weights, cutoff, d, connector)

    print("COMPILATION TIME:", time.time() - start_time)

    print(
        "GRAPH SIZE:",
        measure_graph_size(
            enhanced_calculate_mean_position, weights, cutoff, d, connector
        ),
    )

    sum_ = 0.0

    for i in range(NUMBER_OF_ITERATIONS):
        weights = tf.Variable(
            pq.cvqnn.generate_random_cvqnn_weights(layer_count=layer_count, d=d)
        )

        start_time = time.time()
        result = enhanced_calculate_mean_position(weights, cutoff, d, connector)
        end_time = time.time()

        runtime = end_time - start_time

        print(f"{i}. runtime={runtime},\t result={result[0].numpy()}")

        sum_ += runtime

    print("AVERAGE RUNTIME:", sum_ / NUMBER_OF_ITERATIONS)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest


pytestmark = pytest.mark.benchmark(
    group="gaussian-preparation",
)


def simple_piquasso_benchmark(
    benchmark, pq_gaussian_simulator, example_gaussian_pq_program
):
    @benchmark
    def _():
        pq_gaussian_simulator.execute(example_gaussian_pq_program)


def simple_strawberryfields_benchmark(
    benchmark, example_gaussian_sf_program_and_engine
):
    program, engine = example_gaussian_sf_program_and_engine

    benchmark(engine.run, program)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple, Dict

import numpy as np

from piquasso.api.config import Config
from piquasso.api.exceptions import InvalidState, PiquassoException
from piquasso.api.connector import BaseConnector

from piquasso._math.fock import cutoff_fock_space_dim, get_fock_space_basis
from piquasso._math.linalg import vector_absolute_square
from piquasso._math.indices import (
    get_index_in_fock_space,
    get_index_in_fock_space_array,
    get_auxiliary_modes,
)

from ..state import BaseFockState
from ..general.state import FockState


class PureFockState(BaseFockState):
    r"""A simulated pure Fock state.

    If no mixed states are needed for a Fock simulation, then this state is the most
    appropriate currently, since it does not store an entire density matrix, only a
    state vector with size

    .. math::
        {d + c - 1 \choose c - 1},

    where :math:`c \in \mathbb{N}` is the Fock space cutoff.

    :ivar state_vector: The state vector of the quantum state.
    """

    def __init__(
        self, *, d: int, connector: BaseConnector, config: Optional[Config] = None
    ) -> None:
        """
        Args:
            d (int): The number of modes.
            connector (BaseConnector): Instance containing calculation functions.
            config (Config): Instance containing constants for the simulation.
        """

        super().__init__(d=d, connector=connector, config=config)

        self.state_vector = self._get_empty()

    def _get_empty_list(self) -> list:
        state_vector_size = cutoff_fock_space_dim(cutoff=self._config.cutoff, d=self.d)
        return [0.0] * state_vector_size

    def _get_empty(self) -> np.ndarray:
        state_vector_size = cutoff_fock_space_dim(cutoff=self._config.cutoff, d=self.d)

        return self._np.zeros(
            shape=(state_vector_size,), dtype=self._config.complex_dtype
        )

    def reset(self) -> None:
        state_vector_list = self._get_empty_list()
        state_vector_list[0] = 1.0

        self.state_vector = self._np.array(
            state_vector_list, dtype=self._config.complex_dtype
        )

    def _nonzero_elements_for_single_state_vector(self, state_vector):
        np = self._connector.np
        nonzero_indices = np.nonzero(state_vector)[0]

        occupation_numbers = self._space[nonzero_indices]

        nonzero_elements = state_vector[nonzero_indices]

        for index, coefficient in enumerate(nonzero_elements):
            yield coefficient, tuple(occupation_numbers[index])

    @property
    def nonzero_elements(self):
        return self._nonzero_elements_for_single_state_vector(self.state_vector)

    def _get_repr_for_single_state_vector(self, nonzero_elements):
        return " + ".join(
            [str(coefficient) + str(basis) for coefficient, basis in nonzero_elements]
        )

    def __str__(self) -> str:
        return self._get_repr_for_single_state_vector(self.nonzero_elements)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PureFockState):
            return False
        return self._np.allclose(self.state_vector, other.state_vector)

    @property
    def density_matrix(self) -> np.ndarray:
        cardinality = cutoff_fock_space_dim(d=self.d, cutoff=self._config.cutoff)

        state_vector = self.state_vector[:cardinality]

        return self._np.outer(state_vector, self._np.conj(state_vector))

    def _as_mixed(self) -> FockState:
        return FockState.from_fock_state(self)

    def reduced(self, modes: Tuple[int, ...]) -> FockState:
        return self._as_mixed().reduced(modes)

    def get_particle_detection_probability(
        self, occupation_number: np.ndarray
    ) -> float:
        if self._config.validate and len(occupation_number) != self.d:
            raise PiquassoException(
                f"The specified occupation number should have length '{self.d}': "
                f"occupation_number='{occupation_number}'."
            )

        index = get_index_in_fock_space(occupation_number)

        return self._np.real(
            self.state_vector[index].conjugate() * self.state_vector[index]
        )

    def get_particle_detection_probability_on_modes(
        self,
        occupation_numbers: np.ndarray,
        modes: Tuple[int, ...],
    ) -> float:
        np = self._connector.np
        fallback_np = self._connector.fallback_np

        occupation_numbers = fallback_np.array(occupation_numbers)

        auxiliary_d = self.d - len(modes)
        auxiliary_cutoff = self._config.cutoff - sum(occupation_numbers)
        auxiliary_modes = get_auxiliary_modes(self.d, modes)

        unordered_modes = fallback_np.concatenate([modes, auxiliary_modes])

        card = cutoff_fock_space_dim(d=auxiliary_d, cutoff=auxiliary_cutoff)
        auxiliary_basis = get_fock_space_basis(d=auxiliary_d, cutoff=auxiliary_cutoff)

        repeated_occupation_numbers = fallback_np.repeat(
            occupation_numbers[None, :], card, axis=0
        )

        unordered_occupation_numbers = fallback_np.concatenate(
            [repeated_occupation_numbers, auxiliary_basis], axis=1
        )

        sorter = fallback_np.argsort(unordered_modes)

        ordered_occupation_numbers = unordered_occupation_numbers[:, sorter]

        indices = get_index_in_fock_space_array(ordered_occupation_numbers)

        return np.sum(np.abs(self.state_vector[indices]) ** 2)

    @property
    def fock_probabilities(self) -> np.ndarray:
        return vector_absolute_square(self.state_vector, self._connector)

    @property
    def fock_probabilities_map(self) -> Dict[Tuple[int, ...], float]:
        probability_map: Dict[Tuple[int, ...], float] = {}

        for index, basis in enumerate(self._space):
            probability_map[tuple(basis)] = self._np.abs(self.state_vector[index]) ** 2

        return probability_map

    def normalize(self) -> None:
        norm = self.norm

        if self._config.validate and np.isclose(norm, 0):
            raise InvalidState("The norm of the state is 0.")

        self.state_vector = self.state_vector / self._np.sqrt(norm)

    def validate(self) -> None:
        """Validates the represented state.

        Raises:
            InvalidState:
                Raised, if the norm of the state vector is not close to 1.0.
        """
        if not self._config.validate:
            return

        sum_of_probabilities = sum(self.fock_probabilities)

        if not self._np.isclose(sum_of_probabilities, 1.0):
            raise InvalidState(
                f"The sum of probabilities is {sum_of_probabilities}, "
                "instead of 1.0:\n"
                f"fock_probabilities={self.fock_probabilities}"
            )

    def _get_mean_position_indices(self, mode):
        fallback_np = self._connector.fallback_np

        self._space[:, mode] -= 1
        lowered_indices = get_index_in_fock_space_array(self._space)
        self._space[:, mode] += 2
        raised_indices = get_index_in_fock_space_array(self._space)
        self._space[:, mode] -= 1

        relevant_column = self._space[:, mode]

        nonzero_indices_on_mode = (relevant_column > 0).nonzero()[0]
        upper_index = cutoff_fock_space_dim(d=self.d, cutoff=self._config.cutoff - 1)

        multipliers = fallback_np.sqrt(
            fallback_np.concatenate(
                [
                    relevant_column[nonzero_indices_on_mode],
                    relevant_column[:upper_index] + 1,
                ]
            )
        )
        left_indices = fallback_np.concatenate(
            [lowered_indices[nonzero_indices_on_mode], raised_indices[:upper_index]]
        )
        right_indices = fallback_np.concatenate(
            [nonzero_indices_on_mode, fallback_np.arange(upper_index)]
        )

        return multipliers, left_indices, right_indices

    def mean_position(self, mode: int) -> np.ndarray:
        np = self._connector.np
        fallback_np = self._connector.fallback_np
        multipliers, left_indices, right_indices = self._get_mean_position_indices(mode)

        state_vector = self.state_vector

        accumulator = np.dot(
            (multipliers * state_vector[left_indices]),
            state_vector[right_indices],
        )

        return np.real(accumulator) * fallback_np.sqrt(self._config.hbar / 2)

    def quadratures_mean_variance(
        self, modes: Tuple[int, ...], phi: float = 0
    ) -> Tuple[float, float]:
        r"""This method calculates the mean and the variance of the qudrature operators
        for a single qumode state.
        The quadrature operators :math:`x` and :math:`p` for a mode :math:`i`
        can be calculated using the creation and annihilation operators as follows:

        .. math::
            x_i &= \sqrt{\frac{\hbar}{2}} (a_i + a_i^\dagger) \\
                p_i &= -i \sqrt{\frac{\hbar}{2}} (a_i - a_i^\dagger).

        Let :math:`\phi \in [ 0, 2 \pi )`, we can rotate the quadratures using the
        following transformation:

        .. math::
            Q_{i, \phi} = \cos\phi~x_i + \sin\phi~p_i.

        The expectation value :math:`\langle Q_{i, \phi}\rangle` can be calculated as:

        .. math::
            \operatorname{Tr}(\rho_i Q_{i, \phi}),

        where :math:`\rho_i` is the reduced density matrix of the mode :math:`i` and
        :math:`Q_{i, \phi}` is the rotated quadrature operator for a single mode and
        the variance is calculated as:

        .. math::
            \operatorname{\textit{Var}(Q_{i,\phi})} = \langle Q_{i, \phi}^{2}\rangle
                - \langle Q_{i, \phi}\rangle^{2}.

        Args:
            phi (float): The rotation angle. By default it is `0` which means that
                the mean of the position operator is being calculated. For :math:`\phi=
                \frac{\pi}{2}` the mean of the momentum operator is being calculated.
            modes (tuple[int]): The correspoding mode at which the mean of the
                quadratures are being calculated.
        Returns:
            (float, float): A tuple that contains the expectation value and the
                varianceof of the quadrature operator respectively.
        """
        np = self._connector.np

        reduced_dm = self.reduced(modes=modes).density_matrix
        annih = np.diag(np.sqrt(np.arange(1, self._config.cutoff)), 1)
        create = annih.T
        position = (create + annih) * np.sqrt(self._config.hbar / 2)
        momentum = -1j * (annih - create) * np.sqrt(self._config.hbar / 2)

        if phi != 0:
            rotated_quadratures = position * np.cos(phi) + momentum * np.sin(phi)
        else:
            rotated_quadratures = position

        expectation = np.real(np.trace(np.dot(reduced_dm, rotated_quadratures)))
        variance = (
            np.real(
                np.trace(
                    np.dot(reduced_dm, np.dot(rotated_quadratures, rotated_quadratures))
                )
            )
            - expectation**2
        )
        return expectation, variance

    def mean_photon_number(self):
        r"""Returns the mean photon number

        .. math::
            \mathbb{E}_{\ket{\psi}}(\hat{n}) :=
                \bra{\psi} \hat{n} \ket{\psi},

        where :math:`\hat{n}` is the total photon number operator.
        """

        numbers = np.sum(self._space, axis=1)

        return numbers @ (self._np.abs(self.state_vector) ** 2)

    def variance_photon_number(self):
        r"""Returns the photon number variance

        .. math::
            \operatorname{Var}_{\ket{\psi}}(\hat{n}) :=
                \bra{\psi} (\hat{n} - \bar{n})^2 \ket{\psi},

        where :math:`\hat{n}` is the total photon number operator and :math:`\bar{n}`
        is its expectation value given by :meth:`mean_photon_number`.
        """

        numbers = np.sum(self._space, axis=1)

        probabilities = self._np.abs(self.state_vector) ** 2

        mean = numbers @ probabilities

        return (numbers - mean) ** 2 @ probabilities

    def get_tensor_representation(self):
        cutoff = self._config.cutoff
        d = self.d

        return self._connector.scatter(
            self._space,
            self.state_vector,
            [cutoff] * d,
        )

    def copy(self) -> "PureFockState":
        # NOTE: `__deepcopy__` is not allowed for tensorflow variables, so we have to
        # do it explicitely here.
        state = self.__class__(
            d=self.d, connector=self._connector, config=self._config.copy()
        )

        state.state_vector = self._connector.np.copy(self.state_vector)

        return state

    def get_purity(self):
        return 1.0



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import numpy as np

import piquasso as pq

from scipy.linalg import polar, sinhm, coshm, expm


@pytest.fixture
def generate_random_gaussian_state():
    def func(*, d):
        with pq.Program() as random_state_generator:
            for i in range(d):
                alpha = 2 * (np.random.rand() + 1j * np.random.rand()) - 1
                displacement_gate = pq.Displacement(
                    r=np.abs(alpha), phi=np.angle(alpha)
                )

                r = 2 * np.random.rand() - 1
                phi = np.random.rand() * 2 * np.pi
                squeezing_gate = pq.Squeezing(r, phi=phi)

                pq.Q(i) | displacement_gate | squeezing_gate

        simulator = pq.GaussianSimulator(d=d)
        state = simulator.execute(random_state_generator).state
        state.validate()

        return state

    return func


def test_squeezing(state, gaussian_state_assets):
    r = -0.6
    phi = 0.7

    with pq.Program() as program:
        pq.Q(1) | pq.Squeezing(r, phi)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_phaseshift(state, gaussian_state_assets):
    with pq.Program() as program:
        pq.Q(0) | pq.Phaseshifter(phi=np.pi / 3)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_phaseshift_modes_are_shifted_from_original(state):
    original = state.copy()

    angle = np.pi / 3

    with pq.Program() as program:
        pq.Q(0) | pq.Phaseshifter(phi=angle)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    assert state.reduced((0,)) == original.rotated(-angle).reduced((0,))
    assert state.reduced((1, 2)) == original.reduced((1, 2))


def test_beamsplitter(state, gaussian_state_assets):
    theta = np.pi / 4
    phi = 2 * np.pi / 3

    with pq.Program() as program:
        pq.Q(0, 1) | pq.Beamsplitter(theta=theta, phi=phi)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_displacement_with_alpha(state, gaussian_state_assets):
    alpha = 3 - 4j

    with pq.Program() as program:
        pq.Q(1) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_displacement_with_r_and_phi(state, gaussian_state_assets):
    r = 5
    phi = np.pi / 4

    with pq.Program() as program:
        pq.Q(1) | pq.Displacement(r=r, phi=phi)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_displacement_on_multiple_modes(state, gaussian_state_assets):
    alpha = 3 - 4j

    with pq.Program() as program:
        for i in [0, 1]:
            pq.Q(i) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_displacement_on_all_modes(state, gaussian_state_assets):
    alpha = 3 - 4j

    with pq.Program() as program:
        for i in range(state.d):
            pq.Q(i) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_position_displacement(state, gaussian_state_assets):
    x = 2

    with pq.Program() as program:
        pq.Q(1) | pq.PositionDisplacement(x)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_momentum_displacement(state, gaussian_state_assets):
    p = 2.5

    with pq.Program() as program:
        pq.Q(1) | pq.MomentumDisplacement(p)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_displacement(state, gaussian_state_assets):
    alpha = 2 + 3j
    r = 4
    phi = np.pi / 3

    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=r, phi=phi)
        pq.Q(1) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))
        pq.Q(2) | pq.Displacement(r=r, phi=phi) | pq.Displacement(
            r=np.abs(alpha), phi=np.angle(alpha)
        )

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_displacement_and_squeezing(state, gaussian_state_assets):
    r = 4
    phi = np.pi / 3

    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=r, phi=phi)
        pq.Q(1) | pq.Squeezing(r=-0.6, phi=0.7)
        pq.Q(2) | pq.Squeezing(r=0.8)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_two_mode_squeezing(state, gaussian_state_assets):
    r = 4
    phi = np.pi / 3

    with pq.Program() as program:
        pq.Q(1, 2) | pq.Squeezing2(r=r, phi=phi)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_controlled_X_gate(state, gaussian_state_assets):
    s = 2

    with pq.Program() as program:
        pq.Q(1, 2) | pq.ControlledX(s=s)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_controlled_Z_gate(state, gaussian_state_assets):
    s = 2

    with pq.Program() as program:
        pq.Q(1, 2) | pq.ControlledZ(s=s)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_fourier(state, gaussian_state_assets):
    with pq.Program() as program:
        pq.Q(0) | pq.Fourier()

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_mach_zehnder(state, gaussian_state_assets):
    int_ = np.pi / 3
    ext = np.pi / 4

    with pq.Program() as program:
        pq.Q(1, 2) | pq.MachZehnder(int_=int_, ext=ext)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_quadratic_phase(state, gaussian_state_assets):
    with pq.Program() as program:
        pq.Q(0) | pq.QuadraticPhase(0)
        pq.Q(1) | pq.QuadraticPhase(2)
        pq.Q(2) | pq.QuadraticPhase(1)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_displacement_leaves_the_covariance_invariant(state):
    r = 1
    phi = 1

    initial_covariance_matrix = state.xpxp_covariance_matrix

    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=r, phi=phi)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    final_covariance_matrix = state.xpxp_covariance_matrix

    assert np.allclose(initial_covariance_matrix, final_covariance_matrix)


def test_interferometer_for_1_modes(state, gaussian_state_assets):
    alpha = np.exp(1j * np.pi / 3)

    T = np.array(
        [[alpha]],
        dtype=complex,
    )

    with pq.Program() as program:
        pq.Q(1) | pq.Interferometer(T)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_interferometer_for_2_modes(state, gaussian_state_assets):
    theta = np.pi / 4
    phi = np.pi / 3

    random_unitary = np.array(
        [
            [np.cos(theta), np.sin(theta) * np.exp(1j * phi)],
            [-np.sin(theta) * np.exp(-1j * phi), -np.cos(theta)],
        ],
        dtype=complex,
    )

    with pq.Program() as program:
        pq.Q(0, 1) | pq.Interferometer(random_unitary)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_interferometer_for_all_modes(state, gaussian_state_assets):
    self_adjoint = np.array(
        [
            [1, 2j, 3 + 4j],
            [-2j, 2, 5],
            [3 - 4j, 5, 6],
        ],
        dtype=complex,
    )
    unitary = expm(1j * self_adjoint)

    with pq.Program() as program:
        pq.Q(0, 1, 2) | pq.Interferometer(unitary)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_GaussianTransform_for_1_modes(state, gaussian_state_assets):
    r = 0.4

    alpha = np.exp(1j * np.pi / 3)
    beta = np.exp(1j * np.pi / 4)

    passive = np.array(
        [[alpha * np.cosh(r)]],
        dtype=complex,
    )
    active = np.array(
        [[beta * np.sinh(r)]],
        dtype=complex,
    )

    with pq.Program() as program:
        pq.Q(1) | pq.GaussianTransform(passive=passive, active=active)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_GaussianTransform_with_general_squeezing_matrix():
    d = 3

    squeezing_matrix = np.array(
        [
            [0.1, 0.2j, 0.3],
            [0.2j, 0.2, 0.4],
            [0.3, 0.4, 0.1j],
        ],
        dtype=complex,
    )
    U, r = polar(squeezing_matrix)

    theta = np.pi / 3

    global_phase = np.array(
        [
            [np.cos(theta), -np.sin(theta) * np.exp(1j * np.pi / 4), 0],
            [np.sin(theta) * np.exp(-1j * np.pi / 4), np.cos(theta), 0],
            [0, 0, 1],
        ]
    )

    passive = global_phase @ coshm(r)
    active = global_phase @ sinhm(r) @ U.conj()

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(all) | pq.GaussianTransform(passive=passive, active=active)

    simulator = pq.GaussianSimulator(d=d)
    state = simulator.execute(program).state

    state.validate()

    probabilities = state.fock_probabilities

    assert all(probability >= 0 for probability in probabilities)

    assert sum(probabilities) <= 1.0 or np.isclose(sum(probabilities), 1.0)


def test_GaussianTransform_raises_InvalidParameter_for_nonsymplectic_matrix():
    d = 3

    zero_matrix = np.zeros(shape=(d,) * 2)

    with pq.Program():
        pq.Q() | pq.Vacuum()

        with pytest.raises(pq.api.exceptions.InvalidParameter):
            pq.Q(all) | pq.GaussianTransform(passive=zero_matrix, active=zero_matrix)


def test_graph_embedding(state):
    adjacency_matrix = np.array(
        [
            [0, 1, 1],
            [1, 0, 1],
            [1, 1, 0],
        ]
    )

    with pq.Program() as program:
        pq.Q() | pq.Graph(adjacency_matrix)

    simulator = pq.GaussianSimulator(d=state.d)
    state = simulator.execute(program).state
    state.validate()


def test_displacement_leaves_the_covariance_invariant_for_complex_program():
    d = 3

    alphas = [np.exp(1j * np.pi / 4), 1, 1j]

    with pq.Program() as initialization:
        for i, alpha in enumerate(alphas):
            pq.Q(i) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))

        pq.Q(0) | pq.Squeezing(r=1, phi=np.pi / 2)
        pq.Q(1) | pq.Squeezing(r=2, phi=np.pi / 3)
        pq.Q(2) | pq.Squeezing(r=2, phi=np.pi / 4)

    simulator = pq.GaussianSimulator(d=d)
    state = simulator.execute(initialization).state
    state.validate()

    initial_covariance_matrix = state.xpxp_covariance_matrix

    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=1, phi=1)

    simulator = pq.GaussianSimulator(d=d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()

    final_covariance_matrix = state.xpxp_covariance_matrix

    assert np.allclose(initial_covariance_matrix, final_covariance_matrix)


def test_displaced_vacuum_stays_valid():
    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=2, phi=np.pi / 3)

    simulator = pq.GaussianSimulator(d=3)
    state = simulator.execute(program).state
    state.validate()


def test_multiple_displacements_leave_the_covariance_invariant():
    d = 3

    simulator = pq.GaussianSimulator(d=d)

    vacuum = simulator.create_initial_state()

    initial_covariance_matrix = vacuum.xpxp_covariance_matrix

    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=2, phi=np.pi / 3)
        pq.Q(1) | pq.Displacement(r=1, phi=np.pi / 4)
        pq.Q(2) | pq.Displacement(r=1, phi=np.pi / 6)

    state = simulator.execute(program).state
    state.validate()

    assert np.allclose(state.xpxp_covariance_matrix, initial_covariance_matrix)


@pytest.mark.monkey
def test_random_interferometer(generate_random_gaussian_state, generate_unitary_matrix):
    d = 5

    state = generate_random_gaussian_state(d=d)

    T = generate_unitary_matrix(3)

    with pq.Program() as program:
        pq.Q(0, 1, 3) | pq.Interferometer(T)

    simulator = pq.GaussianSimulator(d=d)
    state = simulator.execute(program, initial_state=state).state
    state.validate()


def test_complex_circuit(gaussian_state_assets):
    d = 5

    with pq.Program() as program:
        for i in range(5):
            pq.Q(i) | pq.Squeezing(r=0.1) | pq.Displacement(r=1)

        pq.Q(0, 1) | pq.Beamsplitter(0.0959408065906761, 0.06786053071484363)
        pq.Q(2, 3) | pq.Beamsplitter(0.7730047654405018, 1.453770233324797)
        pq.Q(1, 2) | pq.Beamsplitter(1.0152680371119776, 1.2863559998816205)
        pq.Q(3, 4) | pq.Beamsplitter(1.3205517879465705, 0.5236836466492961)
        pq.Q(0, 1) | pq.Beamsplitter(4.394480318177715, 4.481575657714487)
        pq.Q(2, 3) | pq.Beamsplitter(2.2300919706807534, 1.5073556513699888)
        pq.Q(1, 2) | pq.Beamsplitter(2.2679037068773673, 1.9550229282085838)
        pq.Q(3, 4) | pq.Beamsplitter(3.340269832485504, 3.289367083610399)

        for i in range(5):
            pq.Q(i) | pq.Squeezing(r=0.1) | pq.Displacement(r=1)

        pq.Q(0, 1) | pq.Beamsplitter(0.0959408065906761, 0.06786053071484363)
        pq.Q(2, 3) | pq.Beamsplitter(0.7730047654405018, 1.453770233324797)
        pq.Q(1, 2) | pq.Beamsplitter(1.0152680371119776, 1.2863559998816205)
        pq.Q(3, 4) | pq.Beamsplitter(1.3205517879465705, 0.5236836466492961)
        pq.Q(0, 1) | pq.Beamsplitter(4.394480318177715, 4.481575657714487)
        pq.Q(2, 3) | pq.Beamsplitter(2.2300919706807534, 1.5073556513699888)
        pq.Q(1, 2) | pq.Beamsplitter(2.2679037068773673, 1.9550229282085838)
        pq.Q(3, 4) | pq.Beamsplitter(3.340269832485504, 3.289367083610399)

    simulator = pq.GaussianSimulator(d=d)
    state = simulator.execute(program).state
    state.validate()

    expected_state = gaussian_state_assets.load()
    assert state == expected_state


def test_program_stacking_with_measurement():
    with pq.Program() as preparation:
        pq.Q(0, 1) | pq.Squeezing2(r=1, phi=np.pi / 4)
        pq.Q(2, 3) | pq.Squeezing2(r=2, phi=np.pi / 3)

    with pq.Program() as interferometer:
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4, phi=np.pi / 3)
        pq.Q(1) | pq.Phaseshifter(phi=np.pi / 2)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 6)

    with pq.Program() as program:
        pq.Q(0, 1, 2, 3, 4) | preparation

        pq.Q(0, 1, 2) | interferometer
        pq.Q(2, 3, 4) | interferometer

        pq.Q(3) | pq.HeterodyneMeasurement()

    simulator = pq.GaussianSimulator(d=5)
    state = simulator.execute(program).state
    state.validate()


def test_complex_one_mode_scenario():
    with pq.Program() as program:
        pq.Q(0) | pq.Squeezing(r=np.log(2))
        pq.Q(0) | pq.Displacement(r=1)
        pq.Q(0) | pq.Phaseshifter(np.pi / 4)

    simulator = pq.GaussianSimulator(d=1, config=pq.Config(cutoff=4))
    state = simulator.execute(program).state

    assert np.allclose(
        state.fock_probabilities,
        [
            0.1615172143957243,
            0.41348406885305417,
            0.31024226541130745,
            0.03980473302825398,
        ],
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import piquasso as pq


def test_single_instruction_program_stacking(FakeGate):
    sub_program = pq.Program()
    with sub_program:
        pq.Q(0, 1) | FakeGate(param=420)

    with pq.Program() as program:
        pq.Q(0, 1) | sub_program

    assert program.instructions[0].modes == (0, 1)
    assert program.instructions[0].params == {"param": 420}


def test_multiple_instruction_program_stacking(FakeGate):
    sub_program = pq.Program()
    with sub_program:
        pq.Q(0) | FakeGate(param=2) | FakeGate(param=4)
        pq.Q(2, 3) | FakeGate(param=10)

    with pq.Program() as program:
        pq.Q(0, 1, 2, 3) | sub_program

    assert program.instructions[0].modes == (0,)
    assert program.instructions[0].params == {"param": 2}

    assert program.instructions[1].modes == (0,)
    assert program.instructions[1].params == {"param": 4}

    assert program.instructions[2].modes == (2, 3)
    assert program.instructions[2].params == {"param": 10}


def test_multiple_instruction_mixed_program_stacking(FakeGate):
    sub_program = pq.Program()
    with sub_program:
        pq.Q(0, 1) | FakeGate(param=10)

    with pq.Program() as program:
        pq.Q(2) | FakeGate(param=2)
        pq.Q(0, 1) | sub_program
        pq.Q(3) | FakeGate(param=0)

    assert program.instructions[0].modes == (2,)
    assert program.instructions[0].params == {"param": 2}

    assert program.instructions[1].modes == (0, 1)
    assert program.instructions[1].params == {"param": 10}

    assert program.instructions[2].modes == (3,)
    assert program.instructions[2].params == {"param": 0}


def test_mixed_index_program_stacking(FakeGate):
    sub_program = pq.Program()
    with sub_program:
        pq.Q(0, 1) | FakeGate(param=10)
        pq.Q(2, 3) | FakeGate(param=100)

    with pq.Program() as program:
        pq.Q(0, 2, 1, 3) | sub_program

    assert program.instructions[0].modes == (0, 2)
    assert program.instructions[0].params == {"param": 10}

    assert program.instructions[1].modes == (1, 3)
    assert program.instructions[1].params == {"param": 100}


def test_program_stacking_with_nested_definition(FakeGate):

    def create_subprogram():
        with pq.Program() as sub_program:
            pq.Q(0, 1) | FakeGate(param=1)
            pq.Q(2, 3) | FakeGate(param=2)

        return sub_program

    with pq.Program() as program:
        pq.Q(0, 1) | FakeGate(param=0)
        pq.Q(0, 2, 1, 3) | create_subprogram()
        pq.Q(0, 3) | FakeGate(param=3)

    assert program.instructions[0].modes == (0, 1)
    assert program.instructions[0].params == {"param": 0}

    assert program.instructions[1].modes == (0, 2)
    assert program.instructions[1].params == {"param": 1}

    assert program.instructions[2].modes == (1, 3)
    assert program.instructions[2].params == {"param": 2}

    assert program.instructions[3].modes == (0, 3)
    assert program.instructions[3].params == {"param": 3}



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numba as nb
import numpy as np


@nb.njit(cache=True)
def calculate_loop_corrections(cx_diag_elements, diag_elements, AZ, num_of_modes):
    """
    Calculates loop corrections.
    """
    loop_correction = np.zeros(num_of_modes, dtype=AZ.dtype)

    max_idx = len(cx_diag_elements)

    tmp_vec = np.zeros(max_idx, dtype=AZ.dtype)

    for idx in range(0, num_of_modes):
        tmp = 0.0

        for jdx in range(0, max_idx):
            tmp = tmp + diag_elements[jdx] * cx_diag_elements[jdx]

        loop_correction[idx] = tmp

        tmp = 0.0
        for kdx in range(0, max_idx):
            tmp += AZ[0, kdx] * cx_diag_elements[kdx]

        tmp_vec[0] = tmp

        for jdx in range(1, max_idx):
            tmp = 0.0
            for kdx in range(jdx - 1, max_idx):
                tmp += AZ[jdx, kdx] * cx_diag_elements[kdx]

            tmp_vec[jdx] = tmp

        cx_diag_elements = np.copy(tmp_vec)

    return loop_correction



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq

from jax import grad
import jax.numpy as jnp


def test_Beamsplitter_gradient_at_theta_equal_0():
    connector = pq.JaxConnector()

    def get_fidelity(theta):
        initial_state = [1, 1, 0]

        with pq.Program() as program:
            pq.Q() | pq.StateVector(initial_state)

        with pq.Program() as rotated_program:
            pq.Q() | program

            pq.Q(0, 1) | pq.Beamsplitter(theta=theta)

        simulator = pq.SamplingSimulator(d=3, connector=connector)

        initial_state_vector = simulator.execute(program).state.state_vector
        rotated_state_vector = simulator.execute(rotated_program).state.state_vector

        return jnp.abs(jnp.conj(initial_state_vector) @ rotated_state_vector)

    angle = 0.0

    fidelity_at_0 = get_fidelity(angle)

    assert np.isclose(fidelity_at_0, 1.0)

    grad_get_fidelity = grad(get_fidelity)
    fidelity_grad_at_0 = grad_get_fidelity(angle)

    assert np.isclose(fidelity_grad_at_0, 0.0)


@pytest.mark.monkey
def test_Beamsplitter_gradient_at_random_angle():
    connector = pq.JaxConnector()

    def get_fidelity(theta):
        initial_state = [1, 1, 0]

        with pq.Program() as program:
            pq.Q() | pq.StateVector(initial_state)

        with pq.Program() as rotated_program:
            pq.Q() | program

            pq.Q(0, 1) | pq.Beamsplitter(theta=theta)

        simulator = pq.SamplingSimulator(d=3, connector=connector)

        initial_state_vector = simulator.execute(program).state.state_vector
        rotated_state_vector = simulator.execute(rotated_program).state.state_vector

        return jnp.abs(jnp.conj(initial_state_vector) @ rotated_state_vector)

    angle = np.random.uniform() * 2 * np.pi

    expected_fidelity = np.abs(np.cos(2 * angle))

    fidelity = get_fidelity(angle)

    assert np.isclose(fidelity, expected_fidelity)

    grad_get_fidelity = grad(get_fidelity)
    fidelity_grad = grad_get_fidelity(angle)

    expected_fidelity_grad = -2 * np.sin(2 * angle) * np.sign(np.cos(2 * angle))

    assert np.isclose(fidelity_grad, expected_fidelity_grad)


@pytest.mark.monkey
def test_Beamsplitter_gradient_at_random_angle_multiparticle_initial_state():
    connector = pq.JaxConnector()

    def get_fidelity(theta):
        initial_state = [2, 1, 0]

        with pq.Program() as program:
            pq.Q() | pq.StateVector(initial_state)

        with pq.Program() as rotated_program:
            pq.Q() | program

            pq.Q(0, 1) | pq.Beamsplitter(theta=theta)

        simulator = pq.SamplingSimulator(
            d=3,
            connector=connector,
            config=pq.Config(cutoff=np.sum(initial_state) + 1),
        )

        initial_state_vector = simulator.execute(program).state.state_vector
        rotated_state_vector = simulator.execute(rotated_program).state.state_vector

        return jnp.abs(jnp.conj(initial_state_vector) @ rotated_state_vector)

    angle = np.random.uniform() * 2 * np.pi

    overlap = np.cos(angle) * np.cos(2 * angle) - np.sin(angle) * np.sin(2 * angle) / 2

    expected_fidelity = np.abs(overlap)

    fidelity = get_fidelity(angle)

    assert np.isclose(fidelity, expected_fidelity)

    grad_get_fidelity = grad(get_fidelity)
    fidelity_grad = grad_get_fidelity(angle)

    overlap_grad = -2 * np.sin(angle) * np.cos(2 * angle) - 5 / 2 * np.sin(
        2 * angle
    ) * np.cos(angle)

    expected_fidelity_grad = overlap_grad * (1 if (overlap > 0) else -1)

    assert np.isclose(fidelity_grad, expected_fidelity_grad)


def test_get_particle_detection_probability_gradient():
    theta = jnp.pi / 3

    simulator = pq.SamplingSimulator(
        d=2, config=pq.Config(cutoff=3), connector=pq.JaxConnector()
    )

    def func(theta):
        with pq.Program() as program:
            pq.Q(all) | pq.StateVector((2, 0))

            pq.Q(all) | pq.Beamsplitter(theta=theta)

        state = simulator.execute(program).state

        return state.get_particle_detection_probability(occupation_number=(2, 0))

    get_jacobian = grad(func)

    probability = func(theta)

    assert jnp.isclose(probability, np.cos(theta) ** 4)

    jacobian = get_jacobian(theta)

    assert jnp.isclose(jacobian, -4 * np.cos(theta) ** 3 * np.sin(theta))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np
import piquasso as pq


pytestmark = pytest.mark.benchmark(
    group="gaussian-homodyne-measurement",
)


def piquasso_benchmark(benchmark, pq_gaussian_simulator, example_pq_gaussian_state):
    @benchmark
    def func():
        with pq.Program() as new_program:
            pq.Q(0) | pq.HomodyneMeasurement(phi=np.pi / 4)

            pq_gaussian_simulator.execute(
                new_program, initial_state=example_pq_gaussian_state
            )


def strawberryfields_benchmark(benchmark, example_sf_gaussian_state, d, sf):
    @benchmark
    def func():
        new_program = sf.Program(d)
        new_engine = sf.Engine(backend="gaussian")

        new_program.state = example_sf_gaussian_state

        with new_program.context as q:
            sf.ops.MeasureHomodyne(phi=np.pi / 4) | q[0]

        new_engine.run(new_program)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..simulator import BuiltinSimulator
from piquasso.instructions import preparations, gates, measurements, channels

from piquasso._simulators.connectors import NumpyConnector, JaxConnector

from .state import SamplingState
from .calculations import (
    state_vector,
    passive_linear,
    particle_number_measurement,
    loss,
    lossy_interferometer,
    post_select_photons,
    imperfect_post_select_photons,
)


class SamplingSimulator(BuiltinSimulator):
    """Performs photonic simulations using Fock representation with pure states.

    The simulation (when executed) results in an instance of
    :class:`~piquasso._simulators.sampling.state.SamplingState`.

    Example usage::

        import numpy as np
        import piquasso as pq

        from scipy.stats import unitary_group

        d = 5

        U = unitary_group.rvs(d)

        with pq.Program() as program:
            pq.Q() | pq.StateVector([1, 1, 1, 0, 0])

            pq.Q() | pq.Interferometer(U)

            pq.Q(all) | pq.ParticleNumberMeasurement()

        simulator = pq.SamplingSimulator(d=5)
        result = simulator.execute(program, shots=100)

    Supported preparations:
        :class:`~piquasso.instructions.preparations.StateVector`.

    Supported gates:
        :class:`~piquasso.instructions.gates.Interferometer`,
        :class:`~piquasso.instructions.gates.Beamsplitter`,
        :class:`~piquasso.instructions.gates.Phaseshifter`,
        :class:`~piquasso.instructions.gates.MachZehnder`,
        :class:`~piquasso.instructions.gates.Fourier`,
        :class:`~piquasso.instructions.gates.Interferometer`.

    Supported measurements:
        :class:`~piquasso.instructions.measurements.ParticleNumberMeasurement`.

    Supported channels:
        :class:`~piquasso.instructions.channels.Loss`.
    """

    _instruction_map = {
        preparations.StateVector: state_vector,
        gates.Beamsplitter: passive_linear,
        gates.Beamsplitter5050: passive_linear,
        gates.Phaseshifter: passive_linear,
        gates.MachZehnder: passive_linear,
        gates.Fourier: passive_linear,
        gates.Interferometer: passive_linear,
        measurements.ParticleNumberMeasurement: particle_number_measurement,
        measurements.PostSelectPhotons: post_select_photons,
        measurements.ImperfectPostSelectPhotons: imperfect_post_select_photons,
        channels.Loss: loss,
        channels.LossyInterferometer: lossy_interferometer,
    }

    _state_class = SamplingState

    _default_connector_class = NumpyConnector

    _extra_builtin_connectors = [JaxConnector]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------

project = "Piquasso"
copyright = "2021-2025, Budapest Quantum Computing Group"
author = "Budapest Quantum Computing Group"


# -- General configuration ---------------------------------------------------

add_module_names = False
autodoc_member_order = "bysource"
autoclass_content = "both"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "nbsphinx",
    "sphinxcontrib.bibtex",
    "sphinx_copybutton",
    "sphinx_design",
    "IPython.sphinxext.ipython_console_highlighting",
]


# Intersphinx setup
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "blackbird": ("https://quantum-blackbird.readthedocs.io/en/latest/", None),
    "scipy": ("http://docs.scipy.org/doc/scipy/reference/", None),
    "tensorflow": (
        "https://www.tensorflow.org/api_docs/python",
        "https://github.com/GPflow/tensorflow-intersphinx/raw/master/tf2_py_objects.inv",  # noqa: E501
    ),
}


# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_type_aliases = None

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

html_static_path = ["_static"]
html_favicon = "_static/favicon.png"

pq_color = {
    "pq-grey": "#656565",
    "pq-smokey-white": "#f4f4f4",
    "pq-blue-shade1": "#1e2844",
    "pq-blue-shade2": "#273250",
    "pq-blue-shade3": "#2f3b5a",
    "pq-blue-shade4": "#364469",
    "pq-font-color-shade1": "#ccd1da",
    "pq-font-color-shade2": "#b2bac8",
    "pq-font-color-shade3": "#8995a8",
    "pq-font-color-shade4": "#717c8e",
    "pq-color-red": "#e46363",
    "pq-color-border": "#232d48",
    "pq-color-border-dark": "#1a243e",
}

html_theme_options = {
    "sidebar_hide_name": True,
    "light_logo": "logo_light.svg",
    "dark_logo": "logo_main.svg",
    "dark_css_variables": {
        "color-background-primary": pq_color["pq-blue-shade1"],
        "color-background-secondary": pq_color["pq-blue-shade2"],
        "color-code-background": pq_color["pq-blue-shade3"],
        "color-foreground-primary": pq_color["pq-font-color-shade1"],
        "color-foreground-secondary": pq_color["pq-font-color-shade2"],
        "color-sidebar-link-text--top-level": pq_color["pq-font-color-shade1"],
        "color-brand-primary": pq_color["pq-font-color-shade4"],
        "color-brand-content": pq_color["pq-color-red"],
        "color-admonition-title--note": pq_color["pq-color-red"],
        "color-admonition-background": pq_color["pq-blue-shade3"],
        "color-admonition-title-background--note": pq_color["pq-blue-shade2"],
        "color-background-hover": pq_color["pq-blue-shade3"],
    },
}

html_css_files = ["custom.css"]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".


# NBSPHINX

nbsphinx_prompt_width = "0"

nbsphinx_prolog = r"""
.. note::
    .. raw:: html

        <div>
            You can download this notebook
            <a href="{{ env.docname.split('/')|last|e + '.ipynb' }}" download>here</a>.
        </div>
"""  # TODO: get base path.

bibtex_bibfiles = ["bibliography.bib"]
bibtex_reference_style = "author_year"

nbsphinx_epilog = r"""
.. footbibliography::
"""

# COPYBUTTON

copybutton_exclude = ".linenos, .gp, .go"



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import jax.numpy as jnp

from scipy.linalg import block_diag
from piquasso._math.hafnian import (
    hafnian_with_reduction,
    hafnian_with_reduction_batch,
    loop_hafnian_with_reduction,
    loop_hafnian_with_reduction_batch,
)
from piquasso._math.jax.hafnian import (
    loop_hafnian_with_reduction as jax_loop_hafnian_with_reduction,
)

from piquasso._math.linalg import reduce_


def test_hafnian_with_empty_reduction():
    matrix = np.array(
        [
            [1, -500j],
            [500j, 6],
        ],
        dtype=complex,
    )

    assert np.isclose(hafnian_with_reduction(matrix, np.array([0, 0])), 1.0)


def test_loop_hafnian_with_empty_reduction():
    matrix = np.array(
        [
            [1, -500j],
            [500j, 6],
        ],
        dtype=complex,
    )

    assert np.isclose(
        loop_hafnian_with_reduction(matrix, np.array([0, 0]), np.array([0, 0])), 1.0
    )


def test_hafnian_with_reduction_on_2_by_2_real_matrix():
    matrix = np.array(
        [
            [1, 500],
            [500, 6],
        ],
        dtype=float,
    )

    assert np.isclose(hafnian_with_reduction(matrix, np.array([2, 4])), 18000108)


def test_hafnian_with_reduction_on_2_by_2_complex_matrix():
    matrix = np.array(
        [
            [1, -500j],
            [500j, 6],
        ],
        dtype=complex,
    )

    assert np.isclose(hafnian_with_reduction(matrix, np.array([4, 0])), 3)


def test_loop_hafnian_with_reduction_on_2_by_2_real_matrix():
    matrix = np.array(
        [
            [1, 500],
            [500, 6],
        ],
        dtype=float,
    )

    assert np.isclose(
        loop_hafnian_with_reduction(
            matrix,
            np.array([0.4, 0.6]),
            occupation_numbers=np.array([2, 4]),
        ),
        19097766.06393626,
    )


def test_loop_hafnian_with_reduction_on_1_by_1_complex_matrix():
    matrix = np.array([[1 + 1j]], dtype=complex)

    assert np.isclose(
        loop_hafnian_with_reduction(
            matrix,
            np.array([1j]),
            occupation_numbers=np.array([6]),
        ),
        -16 - 45j,
    )


def test_loop_hafnian_with_reduction_on_2_by_2_complex_matrix():
    matrix = np.array(
        [
            [1, 500j],
            [500j, 6],
        ],
        dtype=complex,
    )

    assert np.isclose(
        loop_hafnian_with_reduction(
            matrix,
            np.array([0.4, 0.6]),
            occupation_numbers=np.array([4, 0]),
        ),
        3.9856,
    )


def test_loop_hafnian_with_reduction_on_2_by_2_complex_matrix_8_particles():
    A = np.array([[1 + 2j, 3 + 4j], [3 + 4j, 5 + 6j]])

    diagonal = np.array([-1 - 2j, 3])

    occupation_numbers = np.array([4, 4])

    assert np.isclose(
        loop_hafnian_with_reduction(A, diagonal, occupation_numbers), 12252 - 5184j
    )


def test_loop_hafnian_with_reduction_on_3_by_3_real_matrix():
    matrix = np.array(
        [
            [1, 2, 3],
            [2, 6, 7],
            [3, 7, 3],
        ],
        dtype=float,
    )

    assert np.isclose(
        loop_hafnian_with_reduction(
            matrix,
            np.array([0.1, 0.2, 0.3]),
            occupation_numbers=np.array([1, 3, 1]),
        ),
        51.28824,
    )


def test_loop_hafnian_with_reduction_on_3_by_3_complex_matrix():
    matrix = np.array(
        [
            [1j, 2, 3j],
            [2, 6, 7],
            [3j, 7, 3],
        ],
        dtype=complex,
    )

    assert np.isclose(
        loop_hafnian_with_reduction(
            matrix,
            np.array([0.1, 0.2j, 0.3]),
            occupation_numbers=np.array([1, 3, 1]),
        ),
        12.468 + 16.90776j,
    )


def test_hafnian_with_reduction_on_4_by_4_real_matrix():
    matrix = np.array(
        [
            [1, 2, 3, 4],
            [2, 6, 7, 8],
            [3, 7, 3, 4],
            [4, 8, 4, 8],
        ],
        dtype=float,
    )

    assert np.isclose(hafnian_with_reduction(matrix, np.array([2, 1, 0, 3])), 1344)


def test_hafnian_with_reduction_on_4_by_4_complex_matrix():
    matrix = np.array(
        [
            [1j, 2, 3j, 4],
            [2, 6, 7, 8j],
            [3j, 7, 3, 4],
            [4, 8j, 4, 8j],
        ],
        dtype=complex,
    )

    assert np.isclose(hafnian_with_reduction(matrix, np.array([2, 1, 0, 3])), 960j)


def test_loop_hafnian_with_reduction_on_4_by_4_real_matrix():
    matrix = np.array(
        [
            [1, 2, 3, 4],
            [2, 6, 7, 8],
            [3, 7, 3, 4],
            [4, 8, 4, 8],
        ],
        dtype=float,
    )

    assert np.isclose(
        loop_hafnian_with_reduction(
            matrix,
            np.array([0.1, 0.2, 0.3, 0.4]),
            occupation_numbers=np.array([2, 1, 0, 4]),
        ),
        3226.0762112,
    )


def test_loop_hafnian_with_reduction_on_4_by_4_complex_matrix():
    matrix = np.array(
        [
            [1j, 2, 3j, 4],
            [2, 6, 7, 8j],
            [3j, 7, 3, 4],
            [4, 8j, 4, 8j],
        ],
        dtype=complex,
    )

    assert np.isclose(
        loop_hafnian_with_reduction(
            matrix,
            np.array([0.1, 0.2j, 0.3, 0.4 + 0.1j]),
            occupation_numbers=np.array([1, 3, 1, 1]),
        ),
        205.376424 + 690.048304j,
    )


def test_hafnian_with_reduction_on_6_by_6_real_matrix():
    matrix = np.array(
        [
            [1, 2, 3, 4, 5, 6],
            [2, 6, 7, 8, 9, 5],
            [3, 7, 3, 4, 3, 7],
            [4, 8, 4, 8, 2, 1],
            [5, 9, 3, 2, 2, 0],
            [6, 5, 7, 1, 0, 1],
        ],
        dtype=float,
    )

    assert np.isclose(hafnian_with_reduction(matrix, np.ones(6, dtype=int)), 1262.0)


def test_hafnian_with_reduction_on_6_by_6_complex_matrix():
    matrix = np.array(
        [
            [1, 2, 3, 4j, 5, 6j],
            [2, 6, 7j, 8, 9, 5],
            [3, 7j, 3, 4, 3, 7],
            [4j, 8, 4, 8, 2, 1],
            [5, 9, 3, 2, 2j, 0],
            [6j, 5, 7, 1, 0, 1],
        ],
        dtype=complex,
    )

    assert np.isclose(
        hafnian_with_reduction(matrix, np.array([1, 1, 1, 1, 0, 2])), -22 + 1192j
    )


def test_loop_hafnian_with_reduction_on_6_by_6_real_matrix():
    matrix = np.array(
        [
            [1, 2, 3, 4, 5, 6],
            [2, 6, 7, 8, 9, 5],
            [3, 7, 3, 4, 3, 7],
            [4, 8, 4, 8, 2, 1],
            [5, 9, 3, 2, 2, 0],
            [6, 5, 7, 1, 0, 1],
        ],
        dtype=float,
    )

    assert np.isclose(
        loop_hafnian_with_reduction(
            matrix,
            np.zeros(6),
            occupation_numbers=np.ones(6, dtype=int),
        ),
        1262,
    )


def test_loop_hafnian_with_reduction_on_6_by_6_complex_matrix():
    matrix = np.array(
        [
            [1, 2, 3, 4j, 5, 6j],
            [2, 6, 7j, 8, 9, 5],
            [3, 7j, 3, 4, 3, 7],
            [4j, 8, 4, 8, 2, 1],
            [5, 9, 3, 2, 2j, 0],
            [6j, 5, 7, 1, 0, 1],
        ],
        dtype=complex,
    )

    assert np.isclose(
        loop_hafnian_with_reduction(
            matrix,
            np.zeros(6),
            occupation_numbers=np.array([1, 1, 1, 1, 0, 2]),
        ),
        -22.0 + 1192.0j,
    )


@pytest.mark.monkey
def test_random_10_by_10_hafnian_with_reduction(generate_symmetric_matrix):
    matrix = generate_symmetric_matrix(10)

    hafnian_with_reduction(matrix, np.random.randint(0, 2, 10))


@pytest.mark.monkey
def test_hafnian_with_reduction_of_complex_symmetric_matrix_with_odd_dimension_is_zero(
    generate_complex_symmetric_matrix,
):
    matrix = generate_complex_symmetric_matrix(3)

    assert np.isclose(hafnian_with_reduction(matrix, np.ones(3, dtype=int)), 0.0)


@pytest.mark.monkey
def test_loop_hafnian_with_reduction_of_complex_symmetric_block_diagonal_matrix(
    generate_complex_symmetric_matrix,
):
    occupation_numbers = np.random.randint(0, 2, 5)

    submatrix = generate_complex_symmetric_matrix(5)
    submatrix_loop_hafnian = loop_hafnian_with_reduction(
        submatrix, np.zeros(5), occupation_numbers
    )

    matrix = block_diag(submatrix, np.conj(submatrix))
    matrix_loop_hafnian = loop_hafnian_with_reduction(
        matrix, np.zeros(10), np.concatenate([occupation_numbers, occupation_numbers])
    )

    assert np.isclose(
        np.conj(submatrix_loop_hafnian) * submatrix_loop_hafnian,
        matrix_loop_hafnian,
    )


def test_hafnian_with_reduction_6_particles():
    A = np.array(
        [
            [0.06918447, 0.50455312, 0.52623749, 0.51119496],
            [0.56996393, 0.62138659, 0.42424946, 0.6582932],
            [0.79820529, 0.36861047, 0.91219376, 0.18621944],
            [0.37366373, 0.3069308, 0.29075932, 0.60748435],
        ]
    )

    A = A + A.T

    haf = hafnian_with_reduction(A, np.array([2, 0, 1, 3]))

    assert np.isclose(haf, 11.024591504142506)


def test_hafnian_with_reduction_12_particles():
    A = np.array(
        [
            [0.06918447, 0.50455312, 0.52623749, 0.51119496],
            [0.56996393, 0.62138659, 0.42424946, 0.6582932],
            [0.79820529, 0.36861047, 0.91219376, 0.18621944],
            [0.37366373, 0.3069308, 0.29075932, 0.60748435],
        ]
    )

    A = A + A.T

    haf = hafnian_with_reduction(A, np.array([6, 3, 2, 3]))

    assert np.isclose(haf, 37655.53201446679)


def test_hafnian_with_reduction_12_particles_with_0():
    A = np.array(
        [
            [0.06918447, 0.50455312, 0.52623749, 0.51119496],
            [0.56996393, 0.62138659, 0.42424946, 0.6582932],
            [0.79820529, 0.36861047, 0.91219376, 0.18621944],
            [0.37366373, 0.3069308, 0.29075932, 0.60748435],
        ]
    )

    A = A + A.T

    haf = hafnian_with_reduction(A, np.array([7, 4, 0, 3]))

    assert np.isclose(haf, 13629.353597466445)


@pytest.mark.monkey
def test_hafnian_with_reduction_equivalence():
    for _ in range(10):
        d = 5
        max_photons = 4
        A = np.random.rand(d, d) + 1j * np.random.rand(d, d)

        A = A + A.T

        occupation_numbers = np.random.randint(0, max_photons, d)

        expected = hafnian_with_reduction(
            reduce_(A, occupation_numbers),
            np.ones(sum(occupation_numbers), dtype=int),
        )
        actual = hafnian_with_reduction(A, occupation_numbers)

        assert np.isclose(expected, actual)


@pytest.mark.monkey
def test_hafnian_with_reduction_scaling_equivalence():
    for _ in range(10):
        d = np.random.randint(1, 10)
        max_photons = np.random.randint(1, 10)
        A = np.random.rand(d, d) + 1j * np.random.rand(d, d)

        A = A + A.T

        occupation_numbers = np.random.randint(0, max_photons, d)

        scaling_factor = np.random.rand() + 1 / 2

        scaled_before = hafnian_with_reduction(scaling_factor * A, occupation_numbers)
        scaled_after = scaling_factor ** (
            np.sum(occupation_numbers) // 2
        ) * hafnian_with_reduction(A, occupation_numbers)

        assert np.isclose(scaled_before, scaled_after)


@pytest.mark.monkey
def test_loop_hafnian_with_reduction_equivalence():
    for _ in range(100):
        d = 5
        max_photons = 6
        A = np.random.rand(d, d) + 1j * np.random.rand(d, d)

        A = A + A.T

        occupation_numbers = np.random.randint(0, max_photons, d)

        diagonal = np.random.rand(d) + 1j * np.random.rand(d)

        reduced_A = reduce_(A, occupation_numbers)

        reduced_diagonal = reduce_(diagonal, occupation_numbers)

        np.fill_diagonal(reduced_A, reduced_diagonal)

        expected = loop_hafnian_with_reduction(
            reduced_A,
            np.diag(reduced_A),
            np.ones(sum(occupation_numbers), dtype=int),
        )
        actual = loop_hafnian_with_reduction(A, diagonal, occupation_numbers)

        assert np.isclose(expected, actual)


@pytest.mark.monkey
def test_loop_hafnian_with_reduction_scaling_equivalence():
    for _ in range(10):
        d = np.random.randint(1, 10)
        max_photons = np.random.randint(1, 5)
        A = np.random.rand(d, d) + 1j * np.random.rand(d, d)

        A = A + A.T

        diagonal = np.random.rand(d) + 1j * np.random.rand(d)

        occupation_numbers = 2 * np.random.randint(0, max_photons, d)

        scaling_factor = np.random.rand() + 1 / 2

        scaled_before = loop_hafnian_with_reduction(
            scaling_factor * A, np.sqrt(scaling_factor) * diagonal, occupation_numbers
        )
        scaled_after = scaling_factor ** (
            np.sum(occupation_numbers) // 2
        ) * loop_hafnian_with_reduction(A, diagonal, occupation_numbers)

        assert np.isclose(scaled_before, scaled_after)


@pytest.mark.monkey
def test_loop_hafnian_with_reduction_odd_even_equivalence():
    for _ in range(100):
        d = np.random.randint(1, 10)
        max_photons = np.random.randint(1, 5)
        A = np.random.rand(d, d) + 1j * np.random.rand(d, d)

        A_odd = A + A.T

        diagonal_odd = np.random.rand(d) + 1j * np.random.rand(d)

        occupation_numbers_odd = 2 * np.random.randint(0, max_photons, d)
        occupation_numbers_odd[0] += 1

        odd_case = loop_hafnian_with_reduction(
            A_odd, diagonal_odd, occupation_numbers_odd
        )

        occupation_numbers_even = np.concatenate([[1], occupation_numbers_odd])
        diagonal_even = np.concatenate([[1.0], diagonal_odd])

        A_even = block_diag([1.0], A_odd)

        even_case = loop_hafnian_with_reduction(
            A_even, diagonal_even, occupation_numbers_even
        )

        assert np.isclose(odd_case, even_case)


@pytest.mark.parametrize(
    "occupation_numbers, cutoff",
    [
        (np.array([1, 3, 0, 0]), 5),
        (np.array([1, 3, 2, 0]), 6),
        (np.array([1, 5, 2, 0]), 9),
        (np.array([1, 5, 2, 0]), 11),
        (np.array([2, 0, 2, 0]), 12),
        (np.array([0, 0, 0, 0]), 7),
        (np.array([2, 2, 4, 0]), 6),
        (np.array([0, 1, 0, 0]), 8),
        (np.array([2, 1, 0, 0]), 7),
        (np.array([2, 1, 4, 0]), 12),
    ],
)
def test_hafnian_with_reduction_batch(occupation_numbers, cutoff):
    matrix = np.array(
        [
            [1j, 2, 3j, 4],
            [2, 6, 7, 8j],
            [3j, 7, 3, 4],
            [4, 8j, 4, 8j],
        ],
        dtype=complex,
    )
    mask = np.array([0, 0, 0, 1], dtype=int)

    assert np.allclose(
        hafnian_with_reduction_batch(matrix, occupation_numbers, cutoff),
        [
            hafnian_with_reduction(matrix, occupation_numbers + i * mask)
            for i in range(cutoff)
        ],
    )


@pytest.mark.monkey
def test_hafnian_with_reduction_batch_random():
    for _ in range(100):
        d = np.random.randint(1, 10)
        max_photons = np.random.randint(1, 5)
        A = np.random.rand(d, d) + 1j * np.random.rand(d, d)

        A = A + A.T

        occupation_numbers = np.random.randint(0, max_photons, d)
        occupation_numbers[-1] = 0

        cutoff = np.random.randint(1, 12)

        mask = np.zeros_like(occupation_numbers)
        mask[-1] = 1

        assert np.allclose(
            hafnian_with_reduction_batch(A, occupation_numbers, cutoff),
            [
                hafnian_with_reduction(A, occupation_numbers + i * mask)
                for i in range(cutoff)
            ],
        )


@pytest.mark.parametrize(
    "occupation_numbers, cutoff",
    [
        (np.array([1, 3, 0, 0]), 5),
        (np.array([1, 3, 2, 0]), 6),
        (np.array([1, 5, 2, 0]), 9),
        (np.array([1, 5, 2, 0]), 11),
        (np.array([2, 0, 2, 0]), 12),
        (np.array([0, 0, 0, 0]), 7),
        (np.array([2, 2, 4, 0]), 6),
        (np.array([0, 1, 0, 0]), 8),
        (np.array([2, 1, 0, 0]), 7),
        (np.array([2, 1, 4, 0]), 12),
        (np.array([2, 4, 5, 0]), 16),
        (np.array([1, 1, 1, 0]), 15),
    ],
)
def test_loop_hafnian_with_reduction_batch(occupation_numbers, cutoff):
    matrix = np.array(
        [
            [1j, 2, 3j, 4],
            [2, 6, 7, 8j],
            [3j, 7, 3, 4],
            [4, 8j, 4, 8j],
        ],
        dtype=complex,
    )
    mask = np.array([0, 0, 0, 1], dtype=int)
    diagonal = np.array([0.1, 0.2, 0.3, 0.4], dtype=complex)

    assert np.allclose(
        loop_hafnian_with_reduction_batch(matrix, diagonal, occupation_numbers, cutoff),
        [
            loop_hafnian_with_reduction(matrix, diagonal, occupation_numbers + i * mask)
            for i in range(cutoff)
        ],
    )


@pytest.mark.monkey
def test_loop_hafnian_with_reduction_batch_random():
    for _ in range(100):
        d = np.random.randint(1, 12)
        max_photons = np.random.randint(1, 5)
        A = np.random.rand(d, d) + 1j * np.random.rand(d, d)
        A = A + A.T
        diagonal = np.random.rand(d) + 1j * np.random.rand(d)

        occupation_numbers = np.random.randint(0, max_photons, d)
        occupation_numbers[-1] = 0

        cutoff = np.random.randint(1, 12)

        mask = np.zeros_like(occupation_numbers)
        mask[-1] = 1

        expected = [
            loop_hafnian_with_reduction(A, diagonal, occupation_numbers + i * mask)
            for i in range(cutoff)
        ]
        actual = loop_hafnian_with_reduction_batch(
            A, diagonal, occupation_numbers, cutoff
        )

        assert np.allclose(expected, actual)


def test_jax_loop_hafnian_with_reduction():
    matrix = jnp.array(
        [
            [1j, 2, 3j, 4],
            [2, 6, 7, 8j],
            [3j, 7, 3, 4],
            [4, 8j, 4, 8j],
        ],
        dtype=complex,
    )

    assert jnp.isclose(
        jax_loop_hafnian_with_reduction(
            matrix,
            jnp.array([0.1, 0.2j, 0.3, 0.4 + 0.1j]),
            jnp.array([1, 3, 1, 1]),
        ),
        205.376424 + 690.048304j,
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import builtins

import pytest


@pytest.fixture
def raise_ImportError_when_importing_tensorflow(tf):
    real_import = builtins.__import__

    def fake_import(name, globals, locals, fromlist, level):
        if name == "tensorflow":
            raise ImportError()

        return real_import(name, globals, locals, fromlist, level)

    builtins.__import__ = fake_import

    yield

    builtins.__import__ = real_import


@pytest.fixture
def unimport_tensorflow(tf):
    """
    Deletes `tensorflow` from `sys.modules`.
    """

    if "tensorflow" in sys.modules:
        del sys.modules["tensorflow"]


def test_TensorflowConnector_imports_tensorflow_if_installed(tf):
    import piquasso as pq

    pq.TensorflowConnector()

    assert "tensorflow" in sys.modules


def test_TensorflowConnector_raises_ImportError_if_TensorFlow_not_installed(
    raise_ImportError_when_importing_tensorflow,
    tf,
):
    import piquasso as pq

    with pytest.raises(ImportError) as error:
        pq.TensorflowConnector()

    assert error.value.args[0] == (
        "You have invoked a feature which requires 'tensorflow'.\n"
        "You can install tensorflow via:\n"
        "\n"
        "pip install piquasso[tensorflow]"
    )


def test_importing_Piquasso_does_not_import_tensorflow(unimport_tensorflow):
    import piquasso as pq  # noqa: F401

    assert "tensorflow" not in sys.modules


def test_Piquasso_works_without_TensorFlow(
    unimport_tensorflow,
    raise_ImportError_when_importing_tensorflow,
):
    import piquasso as pq  # noqa: F401

    assert "tensorflow" not in sys.modules



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import glob

from itertools import islice
from pathlib import Path
from datetime import date


COPYRIGHT = f"""\
#
# Copyright 2021-{date.today().year} Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

COPYRIGHT_CPP = f"""\
/*
 * Copyright 2021-{date.today().year} Budapest Quantum Computing Group
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
"""

COPYRIGHT_MAP = {
    "*.py": COPYRIGHT,
    "*.cpp": COPYRIGHT_CPP,
    "*.hpp": COPYRIGHT_CPP,
}


def test_copyrights():
    root = Path(__file__).parents[1]

    exceptions = [
        root / ".venv",
        root / "docs",
        root / "dist",
        root / "build",
    ]

    extensions = COPYRIGHT_MAP.keys()

    for extension in extensions:
        copyright_text = COPYRIGHT_MAP[extension]
        for filename in glob.iglob(str(root / "**" / extension), recursive=True):
            if (
                _is_current_file(filename)
                or _is_empty_file(filename)
                or _in_exceptions(filename, exceptions)
            ):
                continue

            with open(filename, "r+") as file:
                first_few_lines = "".join(
                    islice(file.readlines(), copyright_text.count("\n"))
                )

            assert (
                copyright_text in first_few_lines
            ), f"Invalid or no copyright found in {filename}"


def _is_current_file(filename):
    return filename == os.path.abspath(__file__)


def _is_empty_file(filename):
    return os.stat(filename).st_size == 0


def _in_exceptions(filename, exceptions):
    for exception in exceptions:
        if Path(exception) in Path(filename).parents:
            return True



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import piquasso as pq
import tensorflow as tf


pytestmark = pytest.mark.benchmark(
    group="tensorflow",
)


d_tuple = tuple(range(2, 6))


@pytest.mark.parametrize("d, cutoff", zip(d_tuple, d_tuple))
def piquasso_benchmark(benchmark, cutoff, d):
    @benchmark
    def func():
        r = tf.Variable(0.05)
        simulator = pq.PureFockSimulator(
            d=d, config=pq.Config(cutoff=cutoff), connector=pq.TensorflowConnector()
        )

        with tf.GradientTape() as tape:
            with pq.Program() as program:
                pq.Q() | pq.Vacuum()
                for i in range(d):
                    pq.Q(i) | pq.Displacement(r=r)

            state = simulator.execute(program).state
            mean = state.mean_photon_number()

        tape.gradient(mean, [r])


@pytest.mark.parametrize("d, cutoff", zip(d_tuple, d_tuple))
def strawberryfields_benchmark(benchmark, cutoff, d, sf):
    @benchmark
    def func():
        new_program = sf.Program(d)

        mapping = {}

        alpha = tf.Variable(0.05)
        param = new_program.params("alpha")
        mapping["alpha"] = alpha

        new_engine = sf.Engine(backend="tf", backend_options={"cutoff_dim": cutoff})

        with new_program.context as q:
            for i in range(d):
                sf.ops.Dgate(param) | q[i]

        with tf.GradientTape() as tape:
            result = new_engine.run(new_program, args=mapping)
            state = result.state
            mean = sum([state.mean_photon(mode)[0] for mode in range(d)])

        tape.gradient(mean, [alpha])



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import piquasso as pq

import tensorflow as tf
import numpy as np
import time

from piquasso.decompositions.clements import (
    get_weights_from_interferometer,
    get_interferometer_from_weights,
)
from piquasso._math.indices import get_operator_index

from typing import Callable


LR = 0.00025
ITERATIONS = 1000
ALPHA = 10
BETA = 10_000
S_STAR = 0.075

tf.get_logger().setLevel("ERROR")


def get_expected_density_matrix(
    state_vector: tf.Tensor, cutoff: int, connector: pq.TensorflowConnector
) -> tf.Tensor:
    state_00 = [0, 1, 0, 1]
    state_01 = [0, 1, 1, 0]
    state_10 = [1, 0, 0, 1]
    state_11 = [1, 0, 1, 0]

    config = pq.Config(normalize=False, cutoff=cutoff)
    expected_program = pq.Program(
        instructions=[
            pq.StateVector(state_00, coefficient=state_vector[0]),
            pq.StateVector(state_01, coefficient=state_vector[1]),
            pq.StateVector(state_10, coefficient=state_vector[2]),
            pq.StateVector(state_11, coefficient=-state_vector[3]),
        ]
    )

    simulator = pq.PureFockSimulator(d=4, config=config, connector=connector)
    expected_state = simulator.execute(expected_program).state

    return expected_state.density_matrix


def get_initial_weights():
    connector = pq.NumpyConnector()
    modes = (0, 2, 4, 5)

    U = np.array(
        [
            [-1 / 3, -np.sqrt(2) / 3, np.sqrt(2) / 3, 2 / 3],
            [np.sqrt(2) / 3, -1 / 3, -2 / 3, np.sqrt(2) / 3],
            [
                -np.sqrt(3 + np.sqrt(6)) / 3,
                np.sqrt(3 - np.sqrt(6)) / 3,
                -np.sqrt((3 + np.sqrt(6)) / 2) / 3,
                np.sqrt(1 / 6 - 1 / (3 * np.sqrt(6))),
            ],
            [
                -np.sqrt(3 - np.sqrt(6)) / 3,
                -np.sqrt(3 + np.sqrt(6)) / 3,
                -np.sqrt(1 / 6 - 1 / (3 * np.sqrt(6))),
                -np.sqrt((3 + np.sqrt(6)) / 2) / 3,
            ],
        ]
    )

    V = connector.embed_in_identity(U, get_operator_index(modes), 6)

    return get_weights_from_interferometer(V, connector)


def _calculate_loss(
    weights: tf.Tensor,
    expected_density_matrix: tf.Tensor,
    P: tf.Tensor,
    connector: pq.TensorflowConnector,
    state_vector: tf.Tensor,
    cutoff: int,
    alpha: tf.Tensor,
    beta: tf.Tensor,
    S_star: tf.Tensor,
):
    d = 6
    config = pq.Config(cutoff=cutoff, normalize=False)
    np = connector.np

    ancilla_modes = (4, 5)

    state_00 = [0, 1, 0, 1]
    state_01 = [0, 1, 1, 0]
    state_10 = [1, 0, 0, 1]
    state_11 = [1, 0, 1, 0]

    ancilla_state = [1, 1]

    interferometer = get_interferometer_from_weights(
        weights, d, connector, config.complex_dtype
    )

    program = pq.Program(
        instructions=[
            pq.StateVector(state_00 + ancilla_state, coefficient=state_vector[0]),
            pq.StateVector(state_01 + ancilla_state, coefficient=state_vector[1]),
            pq.StateVector(state_10 + ancilla_state, coefficient=state_vector[2]),
            pq.StateVector(state_11 + ancilla_state, coefficient=state_vector[3]),
            pq.Interferometer(interferometer),
            pq.ImperfectPostSelectPhotons(
                postselect_modes=ancilla_modes,
                photon_counts=ancilla_state,
                detector_efficiency_matrix=P,
            ),
        ]
    )

    simulator = pq.PureFockSimulator(d=d, config=config, connector=connector)

    state = simulator.execute(program).state

    density_matrix = state.density_matrix
    success_prob = np.real(np.trace(density_matrix))
    normalized_density_matrix = density_matrix / success_prob

    fidelity = np.real(np.trace(normalized_density_matrix @ expected_density_matrix))
    loss = (
        1
        - np.sqrt(fidelity)
        + alpha * np.log(1 + np.exp(-beta * (success_prob - S_star))) / beta
    )

    return loss, success_prob, fidelity


def train_step(
    weights: tf.Tensor,
    P: tf.Tensor,
    connector: pq.TensorflowConnector,
    expected_density_matrix: tf.Tensor,
    state_vector: tf.Tensor,
    cutoff: int,
    alpha: tf.Tensor,
    beta: tf.Tensor,
    S_star: tf.Tensor,
):
    with tf.GradientTape() as tape:
        loss, success_prob, fidelity = _calculate_loss(
            weights=weights,
            P=P,
            connector=connector,
            expected_density_matrix=expected_density_matrix,
            state_vector=state_vector,
            cutoff=cutoff,
            alpha=alpha,
            beta=beta,
            S_star=S_star,
        )

    grad = tape.gradient(loss, weights)

    return loss, success_prob, fidelity, grad


def train(
    _train_step: Callable,
    weights: tf.Tensor,
    P: tf.Tensor,
    state_vector: tf.Tensor,
    expected_density_matrix: tf.Tensor,
    connector: pq.TensorflowConnector,
    cutoff: int,
):
    opt = tf.keras.optimizers.Adam(learning_rate=LR)

    for _ in range(ITERATIONS):
        loss, success_prob, fidelity, grad = _train_step(
            weights=weights,
            P=P,
            connector=connector,
            expected_density_matrix=expected_density_matrix,
            state_vector=state_vector,
            cutoff=cutoff,
            alpha=ALPHA,
            beta=BETA,
            S_star=S_STAR,
        )

        opt.apply_gradients(zip([grad], [weights]))

    print(f"Loss: {loss}")
    print(f"success probability: {success_prob}")
    print(f"Fidelity: {fidelity}")
    print(f"Weights: {weights.numpy()}")


def main() -> None:
    cutoff = 5

    decorator = tf.function(jit_compile=True)
    connector = pq.TensorflowConnector(decorate_with=decorator)
    np = connector.np
    fallback_np = connector.fallback_np

    P = fallback_np.array(
        [
            [1.0, 0.1050, 0.0110, 0.0012, 0.001, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.8950, 0.2452, 0.0513, 0.0097, 0.0017, 0.0003, 0.0001, 0.0],
            [0.0, 0.0, 0.7438, 0.3770, 0.1304, 0.0384, 0.0104, 0.0027, 0.0007],
            [0.0, 0.0, 0.0, 0.5706, 0.4585, 0.2361, 0.0996, 0.0375, 0.0132],
            [0.0, 0.0, 0.0, 0.0, 0.4013, 0.4672, 0.3346, 0.1907, 0.0952],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.2565, 0.4076, 0.3870, 0.2862],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1476, 0.3066, 0.3724],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0755, 0.1985],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.9338],
        ]
    )

    P = P[:cutoff, :cutoff]
    P = tf.convert_to_tensor(P)

    state_vector = np.sqrt([1 / 4, 1 / 4, 1 / 4, 1 / 4])

    expected_density_matrix = get_expected_density_matrix(
        state_vector=state_vector, cutoff=cutoff, connector=connector
    )
    expected_density_matrix = tf.convert_to_tensor(expected_density_matrix)

    initial_weights = get_initial_weights()

    weights = tf.Variable(initial_weights.copy(), dtype=tf.float64)

    _train_step = decorator(train_step)

    start = time.time()
    train(
        _train_step=_train_step,
        weights=weights,
        P=P,
        state_vector=state_vector,
        expected_density_matrix=expected_density_matrix,
        connector=connector,
        cutoff=cutoff,
    )
    end = time.time()
    print(f"Took {end - start} seconds")


if __name__ == "__main__":
    main()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

import numpy as np

from piquasso.fermionic._utils import _get_fs_fdags


@pytest.fixture
def generate_fermionic_gaussian_hamiltonian(
    generate_hermitian_matrix, generate_skew_symmetric_matrix
):
    def func(d):
        A = generate_hermitian_matrix(d)
        B = generate_skew_symmetric_matrix(d)

        return np.block([[-A.conj(), B], [-B.conj(), A]])

    return func


@pytest.fixture
def generate_passive_fermionic_gaussian_hamiltonian(generate_hermitian_matrix):
    def func(d):
        A = generate_hermitian_matrix(d)
        B = np.zeros_like(A)

        return np.block([[-A.conj(), B], [-B.conj(), A]])

    return func


@pytest.fixture
def get_majorana_operators():

    def func(d):
        fs, fdags = _get_fs_fdags(d)

        ms = []

        for i in range(d):
            ms.append(fs[i] + fdags[i])
            ms.append(-1j * (fs[i] - fdags[i]))

        return ms

    return func


@pytest.fixture
def get_ladder_operators():

    def func(d):
        fs, fdags = _get_fs_fdags(d)

        return fs + fdags

    return func



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import numba as nb


@nb.njit(cache=True)
def arr_comb(n, k):
    n = np.where((n < 0) | (n < k), 0, n)
    prod = np.ones(n.shape, dtype=np.int64)

    for i in range(k):
        prod *= n - i
        prod = prod // (i + 1)

    return prod


@nb.njit(cache=True)
def comb(n, k):
    if n < 0 or k < 0 or n < k:
        return 0

    prod = 1

    k = min(k, n - k)

    for i in range(k):
        prod *= n - i
        prod //= i + 1

    return prod


@nb.njit(cache=True)
def sort_and_get_parity(array):
    n = len(array)
    parity = 1
    for n in range(n - 1, 0, -1):
        for i in range(n):
            if array[i] > array[i + 1]:
                array[i], array[i + 1] = array[i + 1], array[i]
                parity *= -1

    return array, parity


@nb.njit(cache=True)
def partitions(boxes, particles, out=None):
    r"""
    Returns all the possible ways to put a specified number of particles in a specified
    number of boxes in anti-lexicographic order.

    Args:
        boxes: Number of boxes.
        particles: Number of particles.
        out: Optional output array.
    """

    positions = particles + boxes - 1

    if positions == -1 or boxes == 0:
        return np.empty((1, 0), dtype=np.int32)

    size = comb(positions, boxes - 1)

    if out is None:
        result = np.empty((size, boxes), dtype=np.int32)
    else:
        result = out

    separators = np.arange(boxes - 1, dtype=np.int32)
    index = size - 1

    while True:
        prev = -1
        for i in range(boxes - 1):
            result[index, i] = separators[i] - prev - 1
            prev = separators[i]

        result[index, boxes - 1] = positions - prev - 1
        index -= 1

        if index < 0:
            break

        i = boxes - 2
        while separators[i] == positions - (boxes - 1 - i):
            i -= 1

        separators[i] += 1
        for j in range(i + 1, boxes - 1):
            separators[j] = separators[j - 1] + 1

    return result



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Fermionic package
=================

Piquasso package for Fermionic quantum computation.

Note:
    This feature is still experimental.

Basic notations
---------------

Let :math:`f_k` and :math:`f_k^\dagger` denote the Dirac operators. These fulfill the
CAR algebra, i.e.,

.. math::
    \{ f_i, f_j^\dagger \} &= I \delta_{ij}, \\\\
    \{ f_i, f_j \} &= \{ f_i^\dagger, f_j^\dagger \} = 0.

Let us define :math:`\mathbf{f}` as

.. math::
    \mathbf{f} = [f_1^\dagger, \dots, f_d^\dagger, f_1, \dots, f_d].

The Majorana operators are defined as

.. math::

    x_k &:= f_k + f_k^\dagger, \\\\
    p_k &:= -i (f_k - f_k^\dagger),

where :math:`f_k` and :math:`f_k^\dagger` denote the Dirac operators.

:math:`\mathbf{m}` denotes the vector of Majorana operators, i.e.,

.. math::

    \mathbf{m} := [x_1, \dots, x_d, p_1, \dots, p_d].

.. automodule:: piquasso.fermionic.instructions
   :members:
   :show-inheritance:

.. automodule:: piquasso.fermionic.gaussian
   :members:
   :undoc-members:
   :show-inheritance:
   :inherited-members:

.. automodule:: piquasso.fermionic.fock
   :members:
   :undoc-members:
   :show-inheritance:
   :inherited-members:
"""

from . import gaussian

from .gaussian import GaussianState, GaussianSimulator
from .instructions import (
    GaussianHamiltonian,
    ParentHamiltonian,
    ControlledPhase,
    IsingXX,
)
from .fock import PureFockState, PureFockSimulator


__all__ = [
    # Gaussian module
    "gaussian",
    # States
    "GaussianState",
    "PureFockState",
    # Simulators
    "GaussianSimulator",
    # Preparations
    "ParentHamiltonian",
    "PureFockSimulator",
    # Gates
    "GaussianHamiltonian",
    "ControlledPhase",
    "IsingXX",
]



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import piquasso as pq
import numpy as np


pytestmark = pytest.mark.benchmark(
    group="gaussian-threshold-detection",
)


def piquasso_benchmark(benchmark, pq_gaussian_simulator):
    @benchmark
    def func():
        with pq.Program() as program:
            pq.Q(0) | pq.Squeezing(r=1.0, phi=np.pi / 1)
            pq.Q(1) | pq.Squeezing(r=0.2, phi=np.pi / 2)
            pq.Q(2) | pq.Squeezing(r=0.3, phi=np.pi / 3)
            pq.Q(3) | pq.Squeezing(r=0.4, phi=np.pi / 4)
            pq.Q(4) | pq.Squeezing(r=1.0, phi=np.pi / 5)

            pq.Q(0, 1) | pq.Beamsplitter(0.0959408065906761, 0.06786053071484363)
            pq.Q(2, 3) | pq.Beamsplitter(0.7730047654405018, 1.453770233324797)
            pq.Q(1, 2) | pq.Beamsplitter(1.0152680371119776, 1.2863559998816205)
            pq.Q(3, 4) | pq.Beamsplitter(1.3205517879465705, 0.5236836466492961)
            pq.Q(0, 1) | pq.Beamsplitter(4.394480318177715, 4.481575657714487)
            pq.Q(2, 3) | pq.Beamsplitter(2.2300919706807534, 1.5073556513699888)
            pq.Q(1, 2) | pq.Beamsplitter(2.2679037068773673, 1.9550229282085838)
            pq.Q(3, 4) | pq.Beamsplitter(3.340269832485504, 3.289367083610399)

            pq.Q(0, 1, 2) | pq.ThresholdMeasurement()

        pq_gaussian_simulator.execute(program, shots=20)


def strawberryfields_benchmark(benchmark, d, sf):
    @benchmark
    def func():
        program = sf.Program(d)
        engine = sf.Engine(backend="gaussian")

        with program.context as q:
            sf.ops.Sgate(1.0, phi=np.pi / 1) | q[0]
            sf.ops.Sgate(0.2, phi=np.pi / 2) | q[1]
            sf.ops.Sgate(0.3, phi=np.pi / 3) | q[2]
            sf.ops.Sgate(0.4, phi=np.pi / 4) | q[3]
            sf.ops.Sgate(1.0, phi=np.pi / 5) | q[4]

            sf.ops.BSgate(0.0959408065906761, 0.06786053071484363) | (q[0], q[1])
            sf.ops.BSgate(0.7730047654405018, 1.453770233324797) | (q[2], q[3])
            sf.ops.BSgate(1.0152680371119776, 1.2863559998816205) | (q[1], q[2])
            sf.ops.BSgate(1.3205517879465705, 0.5236836466492961) | (q[3], q[4])
            sf.ops.BSgate(4.394480318177715, 4.481575657714487) | (q[0], q[1])
            sf.ops.BSgate(2.2300919706807534, 1.5073556513699888) | (q[2], q[3])
            sf.ops.BSgate(2.2679037068773673, 1.9550229282085838) | (q[1], q[2])
            sf.ops.BSgate(3.340269832485504, 3.289367083610399) | (q[3], q[4])

            sf.ops.MeasureThreshold() | (q[0], q[1], q[2])

        engine.run(program, shots=20)
